import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class MixinUFunc:
    N_test_lipschitz = 10

    @pytest.fixture
    def dim(self):
        return 40

    @pytest.fixture
    def data_shape(self, dim):
        return dim, dim

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        # override in tests if dom(f) != R
        return self._random_array((self.N_test_lipschitz, dim))

    @pytest.fixture
    def data_math_diff_lipschitz(self, dim):
        # override in tests if dom(f) != R
        return self._random_array((self.N_test_lipschitz, dim))


class MixinMapUFunc(MixinUFunc, conftest.MapT):
    pass


class MixinDiffMapUFunc(MixinUFunc, conftest.DiffMapT):
    pass


# Trigonometric Functions =====================================================
class TestSin(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.sin(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(0, 2 * np.pi, dim)
        B = np.sin(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestCos(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.cos(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(0, 2 * np.pi, dim)
        B = np.cos(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestTan(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.tan(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-np.pi, np.pi, dim)
        B = np.tan(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestArcSin(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.arcsin(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-1, 1, dim)
        B = np.arcsin(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        return np.clip(self._random_array((self.N_test_lipschitz, dim)), -1, 1)

    @pytest.fixture
    def data_math_diff_lipschitz(self, dim):
        return np.clip(self._random_array((self.N_test_lipschitz, dim)), -1, 1)


class TestArcCos(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.arccos(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-1, 1, dim)
        B = np.arccos(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        return np.clip(self._random_array((self.N_test_lipschitz, dim)), -1, 1)

    @pytest.fixture
    def data_math_diff_lipschitz(self, dim):
        return np.clip(self._random_array((self.N_test_lipschitz, dim)), -1, 1)


class TestArcTan(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.arctan(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-100, 100, dim)
        B = np.arctan(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


# Hyperbolic Functions ========================================================
class TestSinh(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.sinh(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-3, 3, dim)
        B = np.sinh(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestCosh(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.cosh(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-3, 3, dim)
        B = np.cosh(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestTanh(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.tanh(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-3, 3, dim)
        B = np.tanh(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestArcSinh(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.arcsinh(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-10, 10, dim)
        B = np.arcsinh(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestArcCosh(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.arccosh(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(1, 5, dim)
        B = np.arccosh(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        return np.clip(self._random_array((self.N_test_lipschitz, dim)), a_min=1, a_max=None)

    @pytest.fixture
    def data_math_diff_lipschitz(self, dim):
        return np.clip(self._random_array((self.N_test_lipschitz, dim)), a_min=1, a_max=None)


class TestArcTanh(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.arctanh(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-1, 1, dim)
        B = np.arctanh(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        return np.clip(self._random_array((self.N_test_lipschitz, dim)), a_min=-1, a_max=1)

    @pytest.fixture
    def data_math_diff_lipschitz(self, dim):
        return np.clip(self._random_array((self.N_test_lipschitz, dim)), a_min=-1, a_max=1)


# Exponential Functions =======================================================
class TestExp(MixinDiffMapUFunc):
    @pytest.fixture(params=[None, 2, 10])
    def base(self, request):
        return request.param

    @pytest.fixture
    def spec(self, dim, base, _spec):
        op = pxo.exp(pxo.IdentityOp(dim), base)
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim, base):
        A = np.linspace(-4, 4, dim)
        if base is not None:
            A *= np.log(base)
        B = np.exp(A)

        return dict(
            in_=dict(arr=np.linspace(-4, 4, dim)),
            out=B,
        )


class TestLog(MixinDiffMapUFunc):
    @pytest.fixture(params=[None, 2, 10])
    def base(self, request):
        return request.param

    @pytest.fixture
    def spec(self, dim, base, _spec):
        op = pxo.log(pxo.IdentityOp(dim), base)
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim, base):
        A = np.linspace(0.1, 10, dim)
        B = np.log(A)
        if base is not None:
            B /= np.log(base)
        return dict(
            in_=dict(arr=A),
            out=B,
        )

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        return np.abs(self._random_array((self.N_test_lipschitz, dim)))

    @pytest.fixture
    def data_math_diff_lipschitz(self, dim):
        return np.abs(self._random_array((self.N_test_lipschitz, dim)))


# Miscellaneous Functions =====================================================
class TestClip(MixinMapUFunc):
    @pytest.fixture(params=([0, None], [None, 1], [0, 1]))
    def lims(self, request):
        return request.param

    @pytest.fixture
    def spec(self, dim, lims, _spec):
        op = pxo.clip(pxo.IdentityOp(dim), a_min=lims[0], a_max=lims[1])
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim, lims):
        A = np.linspace(-100, 100, dim)
        B = np.clip(A, a_min=lims[0], a_max=lims[1])
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestSqrt(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.sqrt(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(0, 100, dim)
        B = np.sqrt(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        return np.abs(self._random_array((self.N_test_lipschitz, dim)))

    @pytest.fixture
    def data_math_diff_lipschitz(self, dim):
        return np.abs(self._random_array((self.N_test_lipschitz, dim)))


class TestCbrt(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.cbrt(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-10, 10, dim)
        B = np.cbrt(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestSquare(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.square(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-2, 2, dim)
        B = np.square(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestAbs(MixinMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.abs(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-4, 4, dim)
        B = np.abs(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestSign(MixinMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.sign(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-100, 100, dim)
        B = np.sign(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


# Activation Functions ========================================================
class TestGaussian(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.gaussian(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-4, 4, dim)
        B = np.exp(-(A**2))
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestSigmoid(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.sigmoid(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-4, 4, dim)
        B = 1 / (1 + np.exp(-A))
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestSoftPlus(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.softplus(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-4, 4, dim)
        B = np.log1p(np.exp(A))
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestLeakyReLU(MixinMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.leakyrelu(pxo.IdentityOp(dim), alpha=0.01)
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-4, 4, dim)
        B = np.where(A >= 0, A, A * 0.01)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestReLU(MixinMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.relu(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-4, 4, dim)
        B = A.clip(min=0)
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestSiLU(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.silu(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-4, 4, dim)
        B = A / (1 + np.exp(-A))
        return dict(
            in_=dict(arr=A),
            out=B,
        )


class TestSoftMax(MixinDiffMapUFunc):
    @pytest.fixture
    def spec(self, dim, _spec):
        op = pxo.softmax(pxo.IdentityOp(dim))
        return op, _spec[0], _spec[1]

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(-4, 4, dim)
        exp_A = np.exp(A)
        B = exp_A / np.sum(exp_A, axis=-1)
        return dict(
            in_=dict(arr=A),
            out=B,
        )



# How ArgScaleRule tests work:
#
# * ArgScaleRuleMixin auto-defines all arithmetic method (input,output) pairs.
#   [Caveat: we assume all tested examples are defined on \bR.] (This is not a problem in practice.)
#   [Caveat: we assume the base operators (op_orig) are correctly implemented.] (True if choosing test operators from examples/.)
#
# * To test an arg-scaled-operator, inherit from ArgScaleRuleMixin and the suitable conftest.MapT
#   subclass which the arg-scaled operator should abide by.


import collections.abc as cabc
import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest

op_scale_params = frozenset({-2.3, -1, 1, 2})
op_scale_unit = frozenset(_ for _ in op_scale_params if np.isclose(np.abs(_), 1))
op_scale_nonunit = frozenset(_ for _ in op_scale_params if ~np.isclose(np.abs(_), 1))
op_scale_positive = frozenset(_ for _ in op_scale_params if _ > 0)
op_scale_negative = frozenset(_ for _ in op_scale_params if _ < 0)
op_scale_nonidentity = frozenset(_ for _ in op_scale_params if ~np.isclose(_, 1))


class ArgScaleRuleMixin:
    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def op_orig(self) -> pxt.OpT:
        # Override in inherited class with the operator to be arg-scaled.
        raise NotImplementedError

    @pytest.fixture(params=op_scale_params)
    def op_scale(self, request) -> pxt.Real:
        # Arg-scale factors applied to op_orig()
        return request.param

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, op_orig, op_scale, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        op = op_orig.argscale(op_scale)
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, op_orig) -> pxt.OpShape:
        return op_orig.shape

    @pytest.fixture
    def data_apply(self, op_orig, op_scale) -> conftest.DataLike:
        dim = self._sanitize(op_orig.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        out = op_orig.apply(arr * op_scale)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_adjoint(self, op_orig, op_scale) -> conftest.DataLike:
        codim = self._sanitize(op_orig.codim, 7)
        arr = self._random_array((codim,), seed=20)  # random seed for reproducibility
        out = op_orig.adjoint(arr) * op_scale
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_grad(self, op_orig, op_scale) -> conftest.DataLike:
        dim = self._sanitize(op_orig.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        out = op_orig.grad(arr * op_scale) * op_scale
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_prox(self, op_orig, op_scale) -> conftest.DataLike:
        dim = self._sanitize(op_orig.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        tau = np.abs(self._random_array((1,), seed=21))[0]  # random seed for reproducibility
        out = op_orig.prox(arr * op_scale, tau * (op_scale**2)) / op_scale
        return dict(
            in_=dict(
                arr=arr,
                tau=tau,
            ),
            out=out,
        )

    @pytest.fixture
    def data_math_lipschitz(self, op) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, self._sanitize(op.dim, 7)
        return self._random_array((N_test, dim))

    @pytest.fixture
    def data_math_diff_lipschitz(self, op) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, self._sanitize(op.dim, 7)
        return self._random_array((N_test, dim))

    # Tests -------------------------------------------------------------------
    @pytest.mark.skip("undefined for argscale.")
    def test_interface_asloss(self, op, xp, width):
        pass


# Test classes (Maps) ---------------------------------------------------------
class TestArgScaleRuleMap(ArgScaleRuleMixin, conftest.MapT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_map as tc

        return tc.ReLU(M=6)


class TestArgScaleRuleDiffMap(ArgScaleRuleMixin, conftest.DiffMapT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_diffmap as tc

        return tc.Sin(M=6)


class TestArgScaleRuleLinOp(ArgScaleRuleMixin, conftest.LinOpT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_linop as tc

        return tc.Tile(N=3, M=4)


class TestArgScaleRuleSquareOp(ArgScaleRuleMixin, conftest.SquareOpT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_squareop as tc

        return tc.CumSum(N=7)


class TestArgScaleRuleNormalOp(ArgScaleRuleMixin, conftest.NormalOpT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_normalop as tc

        h = self._random_array((5,), seed=2)
        return tc.CircularConvolution(h=h)


# START UnitOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ArgScaleRuleUnitOp(ArgScaleRuleMixin):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_unitop as tc

        return tc.Permutation(N=7)


@pytest.mark.parametrize("op_scale", op_scale_unit)
class TestArgScaleRuleUnitOp_UnitScale(ArgScaleRuleUnitOp, conftest.UnitOpT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_nonunit)
class TestArgScaleRuleUnitOp_NonUnitScale(ArgScaleRuleUnitOp, conftest.NormalOpT):
    pass


# END UnitOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


class TestArgScaleRuleSelfAdjointOp(ArgScaleRuleMixin, conftest.SelfAdjointOpT):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_selfadjointop as tc

        return tc.SelfAdjointConvolution(N=7)


# START PosDefOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ArgScaleRulePosDefOp(ArgScaleRuleMixin):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_posdefop as tc

        return tc.PSDConvolution(N=7)


@pytest.mark.parametrize("op_scale", op_scale_positive)
class TestArgScaleRulePosDefOp_PositiveScale(ArgScaleRulePosDefOp, conftest.PosDefOpT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_negative)
class TestArgScaleRulePosDefOp_NegativeScale(ArgScaleRulePosDefOp, conftest.SelfAdjointOpT):
    pass


# END PosDefOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


# START ProjOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ArgScaleRuleProjOp(ArgScaleRuleMixin):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_projop as tc

        return tc.Oblique(N=7, alpha=np.pi / 4)


@pytest.mark.parametrize("op_scale", [1])
class TestArgScaleRuleProjOp_IdentityScale(ArgScaleRuleProjOp, conftest.ProjOpT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_nonidentity)
class TestArgScaleRuleProjOp_NonIdentityScale(ArgScaleRuleProjOp, conftest.SquareOpT):
    pass


# END ProjOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


# START OrthProjOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
class ArgScaleRuleOrthProjOp(ArgScaleRuleMixin):
    @pytest.fixture
    def op_orig(self):
        import pyxu_tests.operator.examples.test_orthprojop as tc

        return tc.ScaleDown(N=7)


@pytest.mark.parametrize("op_scale", [1])
class TestArgScaleRuleOrthProjOp_IdentityScale(ArgScaleRuleOrthProjOp, conftest.OrthProjOpT):
    pass


@pytest.mark.parametrize("op_scale", op_scale_nonidentity)
class TestArgScaleRuleOrthProjOp_NonIdentityScale(ArgScaleRuleOrthProjOp, conftest.SelfAdjointOpT):
    pass


# END OrthProjOp ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


# Test classes (Funcs) --------------------------------------------------------
class TestArgScaleRuleFunc(ArgScaleRuleMixin, conftest.FuncT):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_func as tc

        return tc.Median(dim=request.param)


class TestArgScaleRuleDiffFunc(ArgScaleRuleMixin, conftest.DiffFuncT):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_difffunc as tc

        return tc.SquaredL2Norm(M=request.param)


class TestArgScaleRuleProxFunc(ArgScaleRuleMixin, conftest.ProxFuncT):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_proxfunc as tc

        return tc.L1Norm(M=request.param)


class TestArgScaleRuleProxDiffFunc(ArgScaleRuleMixin, conftest.ProxDiffFuncT):
    @pytest.fixture(params=[7])
    def op_orig(self, request):
        import pyxu_tests.operator.examples.test_proxdifffunc as tc

        return tc.SquaredL2Norm(M=request.param)


class TestArgScaleRuleQuadraticFunc(ArgScaleRuleMixin, conftest.QuadraticFuncT):
    @pytest.fixture(params=[0, 1])
    def op_orig(self, request):
        from pyxu_tests.operator.examples.test_linfunc import ScaledSum
        from pyxu_tests.operator.examples.test_posdefop import PSDConvolution

        N = 7
        op = {
            0: pxa.QuadraticFunc(shape=(1, N)),
            1: pxa.QuadraticFunc(
                shape=(1, N),
                Q=PSDConvolution(N=N),
                c=ScaledSum(N=N),
                t=1,
            ),
        }[request.param]
        return op


class TestArgScaleRuleLinFunc(ArgScaleRuleMixin, conftest.LinFuncT):
    @pytest.fixture()
    def op_orig(self):
        import pyxu_tests.operator.examples.test_linfunc as tc

        return tc.ScaledSum(N=7)



import itertools

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest

# Note: Pad.[asarray, pinv, dagger]() are expensive.
# Consider disabling them in test suite.


class TestPad(conftest.LinOpT):
    @pytest.fixture(
        params=[
            # 1D, uni-mode -------------------------------------
            (
                ((5,), 2, "constant"),
                ((5,), ((2, 2),), ("constant",)),
            ),
            (
                ((5,), 2, "wrap"),
                ((5,), ((2, 2),), ("wrap",)),
            ),
            (
                ((5,), 2, "reflect"),
                ((5,), ((2, 2),), ("reflect",)),
            ),
            (
                ((5,), 2, "symmetric"),
                ((5,), ((2, 2),), ("symmetric",)),
            ),
            (
                ((5,), 2, "edge"),
                ((5,), ((2, 2),), ("edge",)),
            ),
            # ND, uni-mode -------------------------------------
            (
                ((5, 3, 4), 2, "constant"),
                ((5, 3, 4), ((2, 2), (2, 2), (2, 2)), ("constant", "constant", "constant")),
            ),
            (
                ((5, 3, 4), 2, "wrap"),
                ((5, 3, 4), ((2, 2), (2, 2), (2, 2)), ("wrap", "wrap", "wrap")),
            ),
            (
                ((5, 3, 4), 2, "reflect"),
                ((5, 3, 4), ((2, 2), (2, 2), (2, 2)), ("reflect", "reflect", "reflect")),
            ),
            (
                ((5, 3, 4), 2, "symmetric"),
                ((5, 3, 4), ((2, 2), (2, 2), (2, 2)), ("symmetric", "symmetric", "symmetric")),
            ),
            (
                ((5, 3, 4), 2, "edge"),
                ((5, 3, 4), ((2, 2), (2, 2), (2, 2)), ("edge", "edge", "edge")),
            ),
            # ND, multi-mode -----------------------------------
            (
                ((5, 3, 4), (2, 1, 3), ("constant", "edge", "wrap")),
                ((5, 3, 4), ((2, 2), (1, 1), (3, 3)), ("constant", "edge", "wrap")),
            ),
            (
                ((5, 3, 4), ((0, 2), (1, 3), (3, 2)), ("constant", "edge", "wrap")),
                ((5, 3, 4), ((0, 2), (1, 3), (3, 2)), ("constant", "edge", "wrap")),
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, pad_width, mode) configs to test.
        # * `request.param[0]` corresponds to raw inputs users provide to Pad().
        # * `request.param[1]` corresponds to their ground-truth canonical parameterization.
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):  # canonical representation
        arg_shape, _, _ = _spec[1]
        return arg_shape

    @pytest.fixture
    def pad_width(self, _spec):  # canonical representation
        _, pad_width, _ = _spec[1]
        return pad_width

    @pytest.fixture
    def mode(self, _spec):  # canonical representation
        _, _, mode = _spec[1]
        return mode

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, _spec, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        arg_shape, pad_width, mode = _spec[0]  # user-provided form
        op = pxo.Pad(
            arg_shape=arg_shape,
            pad_width=pad_width,
            mode=mode,
        )
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, arg_shape, pad_width) -> pxt.OpShape:
        dim = np.prod(arg_shape)

        pad_shape = []
        for N, (lhs, rhs) in zip(arg_shape, pad_width):
            p = N + (lhs + rhs)
            pad_shape.append(p)
        codim = np.prod(pad_shape)

        return (codim, dim)

    @pytest.fixture
    def data_apply(self, arg_shape, pad_width, mode) -> conftest.DataLike:
        arr = self._random_array(arg_shape)
        if len(set(mode)) == 1:  # uni-mode
            out = np.pad(
                array=arr,
                pad_width=pad_width,
                mode=mode[0],
            )
        else:  # multi-mode
            N_dim = len(arg_shape)
            out = arr
            for i in range(N_dim):
                p = [(0, 0)] * N_dim
                p[i] = pad_width[i]
                out = np.pad(
                    array=out,
                    pad_width=p,
                    mode=mode[i],
                )
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )



import collections.abc as cabc
import importlib
import inspect
import types

import numpy as np

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt

__all__ = [
    "copy_if_unsafe",
    "import_module",
    "infer_composition_shape",
    "infer_sum_shape",
    "next_fast_len",
    "parse_params",
    "read_only",
]


def next_fast_len(N: pxt.Integer, even: bool = False) -> pxt.Integer:
    """
    Compute the next `5-smooth number <https://en.wikipedia.org/wiki/Smooth_number>`_ greater-or-equal to `N`.

    If `even=True`, then ensure the next 5-smooth number is even-valued.
    """
    # warning: scipy.fftpack.next_fast_len() != scipy.fft.next_fast_len()
    from scipy.fftpack import next_fast_len

    N5 = next_fast_len(int(N))

    is_even = lambda n: n % 2 == 0
    if (not is_even(N5)) and even:
        while not is_even(N5 := next_fast_len(N5)):
            N5 += 1

    return N5


def peaks(x: pxt.NDArray, y: pxt.NDArray) -> pxt.NDArray:
    r"""
    Matlab 2D peaks function.

    Peaks is a function of two variables, obtained by translating and scaling Gaussian distributions.  (See `Matlab's
    peaks function <https://www.mathworks.com/help/matlab/ref/peaks.html>`_.)

    This function is useful for testing purposes.

    Parameters
    ----------
    x: NDArray
        X coordinates.
    y: NDArray
        Y coordinates.

    Returns
    -------
    z: NDArray
        Values of the 2D function ``peaks`` at the points specified by the entries of `x` and `y`.

    Examples
    --------

    .. plot::

       import numpy as np
       import matplotlib.pyplot as plt
       from pyxu.util.misc import peaks

       x = np.linspace(-3, 3, 1000)
       xx, yy = np.meshgrid(x, x)
       z = peaks(xx, yy)
       plt.figure()
       plt.imshow(z)
    """
    ndi = pxd.NDArrayInfo.from_obj(x)
    xp = ndi.module()

    a = 3 * ((1 - x) ** 2) * xp.exp(-(x**2) - (y + 1) ** 2)
    b = -10 * ((x / 5) - x**3 - y**5) * xp.exp(-(x**2) - (y**2))
    c = -xp.exp(-((x + 1) ** 2) - (y**2)) / 3
    z = a + b + c
    return z


def star_like_sample(
    N: pxt.Integer,
    w: pxt.Integer,
    s: pxt.Real,
    po: pxt.Integer,
    x0: pxt.Real,
    ndi: pxd.NDArrayInfo = pxd.NDArrayInfo.NUMPY,
) -> pxt.NDArray:
    r"""
    Star-like test image.

    Generates a (N, N) square image of a star-like object normalized between 0 and 1.  Based on `GlobalBioIm's
    StarLikeSample function
    <https://github.com/Biomedical-Imaging-Group/GlobalBioIm/blob/master/Util/StarLikeSample.m>`_.  This function is
    useful for testing purposes as it contains high-frequency information.

    Parameters
    ----------
    N: Integer
        Size of the image (must be an even number).
    w: Integer
        The number of branches of the sample will be 4*w.
    s: Real
        Slope of the sigmoid function :math:`\frac1{1+\exp[s (x-x_{0})]}` attenuating the boundaries.
    po: Integer
        Power-raising factor for the final image (to have smoother edges).
    x0: Real
        Radial shift of the sigmoid function :math:`\frac1{1+\exp[s (x-x_{0})]}`.
    ndi: NDArrayInfo
        Desired array module for the output.

    Returns
    -------
    image: NDArray
        (N, N) image of star-like sample.

    Examples
    --------
    .. plot::

       import numpy as np
       import matplotlib.pyplot as plt
       from pyxu.util.misc import star_like_sample

       star = star_like_sample(N=256, w=8, s=20, po=3, x0=0.7)
       plt.figure()
       plt.imshow(star)
    """
    assert N % 2 == 0
    xp = ndi.module()

    grid = xp.linspace(-1, 1, N)
    x, y = xp.meshgrid(grid, grid)
    theta = xp.arctan2(y, x)
    image = 1 + xp.cos(4 * w * theta)
    image /= 1 + xp.exp(s * (xp.sqrt(x**2 + y**2) - x0))
    image = xp.maximum(image, 0) / 2
    image **= po
    image /= xp.amax(image)
    return image


def parse_params(func, *args, **kwargs) -> cabc.Mapping:
    """
    Get function parameterization.

    Returns
    -------
    params: ~collections.abc.Mapping
        (key, value) params as seen in body of `func` when called via `func(*args, **kwargs)`.
    """
    sig = inspect.Signature.from_callable(func)
    f_args = sig.bind(*args, **kwargs)
    f_args.apply_defaults()

    params = dict(
        zip(f_args.arguments.keys(), f_args.args),  # positional arguments
        **f_args.kwargs,
    )
    return params


def import_module(name: str, fail_on_error: bool = True) -> types.ModuleType:
    """
    Load a Python module dynamically.
    """
    try:
        pkg = importlib.import_module(name)
    except ModuleNotFoundError:
        if fail_on_error:
            raise
        else:
            pkg = None
    return pkg


def copy_if_unsafe(x: pxt.NDArray) -> pxt.NDArray:
    """
    Copy array if it is unsafe to do in-place updates on it.

    In-place updates are unsafe if:

    * the array is read-only, OR
    * the array is a view onto another array.

    Parameters
    ----------
    x: NDArray

    Returns
    -------
    y: NDArray
    """
    N = pxd.NDArrayInfo
    ndi = N.from_obj(x)
    if ndi == N.DASK:
        # Dask operations span a graph -> always safe.
        y = x
    elif ndi == N.NUMPY:
        read_only = not x.flags.writeable
        reference = not x.flags.owndata
        y = x.copy() if (read_only or reference) else x
    elif ndi == N.CUPY:
        read_only = False  # https://github.com/cupy/cupy/issues/2616
        reference = not x.flags.owndata
        y = x.copy() if (read_only or reference) else x
    else:
        msg = f"copy_if_unsafe() not yet defined for {ndi}."
        raise NotImplementedError(msg)
    return y


def read_only(x: pxt.NDArray) -> pxt.NDArray:
    """
    Make an array read-only.

    Parameters
    ----------
    x: NDArray

    Returns
    -------
    y: NDArray
    """
    N = pxd.NDArrayInfo
    ndi = N.from_obj(x)
    if ndi == N.DASK:
        # Dask operations are read-only since graph-backed.
        y = x
    elif ndi == N.NUMPY:
        y = x.view()
        y.flags.writeable = False
    elif ndi == N.CUPY:
        y = x.view()
        # y.flags.writeable = False  # https://github.com/cupy/cupy/issues/2616
    else:
        msg = f"read_only() not yet defined for {ndi}."
        raise NotImplementedError(msg)
    return y


def infer_sum_shape(sh1: pxt.OpShape, sh2: pxt.OpShape) -> pxt.OpShape:
    r"""
    Infer shape of arithmetic operation :math:`A + B`.
    """
    A, B, C, D = *sh1, *sh2
    if B == D:
        return np.broadcast_shapes((A,), (C,)) + (B,)
    else:
        raise ValueError(f"Addition of {sh1} and {sh2} operators forbidden.")


def infer_composition_shape(sh1: pxt.OpShape, sh2: pxt.OpShape) -> pxt.OpShape:
    r"""
    Infer shape of arithmetic operation :math:`A \circ B`.
    """
    A, B, C, D = *sh1, *sh2
    if B == C:
        return (A, D)
    else:
        raise ValueError(f"Composition of {sh1} and {sh2} operators forbidden.")



import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "hutchpp",
    "norm",
    "trace",
]


def norm(x: pxt.NDArray, **kwargs):
    """
    This function is identical to :py:func:`numpy.linalg.norm`.  It exists to correct bugs in Dask's implementation.
    """
    xp = pxu.get_array_module(x)
    nrm = xp.linalg.norm(x, **kwargs)
    nrm = nrm.astype(x.dtype, copy=False)  # dask bug: linalg.norm() always returns float64
    return nrm


def trace(
    op: pxa.SquareOp,
    xp: pxt.ArrayModule = None,
    dtype: pxt.DType = None,
) -> pxt.Real:
    r"""
    Exact trace of a linear operator based on multiple evaluation of the forward operator.

    Parameters
    ----------
    op: ~pyxu.abc.operator.SquareOp
    xp: ArrayModule
        Array module used for internal computations. (Default: NumPy.)
    dtype: DType
        Precision to use for internal computations. (Default: current runtime precision.)

    Returns
    -------
    tr: Real
        Exact value of tr(op).
    """
    if xp is None:
        xp = pxd.NDArrayInfo.default().module()

    if dtype is None:
        dtype = pxrt.getPrecision().value
    width = pxrt.Width(dtype)

    tr = 0
    for i in range(op.dim):
        e = xp.zeros(op.dim, dtype=dtype)
        e[i] = 1
        with pxrt.Precision(width):
            tr += op.apply(e)[i]
    return float(tr)


def hutchpp(
    op: pxa.SquareOp,
    m: pxt.Integer = 4002,
    xp: pxt.ArrayModule = None,
    dtype: pxt.DType = None,
    seed: pxt.Integer = None,
) -> pxt.Real:
    r"""
    Stochastic trace estimation of a linear operator based on the Hutch++ algorithm.  (Specifically `algorithm 3 from
    this paper <https://arxiv.org/abs/2010.09649>`_.)

    Parameters
    ----------
    op: ~pyxu.abc.operator.SquareOp
    m: Integer
        Number of queries used to estimate the trace of the linear operator.

        `m` is set to 4002 by default based on the analysis of the variance described in theorem 10.  This default
        corresponds to having an estimation error smaller than 0.01 with probability 0.9.
    xp: ArrayModule
        Array module used for internal computations. (Default: NumPy.)
    dtype: DType
        Precision to use for internal computations. (Default: current runtime precision.)
    seed: Integer
        Seed for the random number generator.

    Returns
    -------
    tr: Real
        Stochastic estimate of tr(op).
    """
    if xp is None:
        xp = pxd.NDArrayInfo.default().module()

    if dtype is None:
        dtype = pxrt.getPrecision().value
    width = pxrt.Width(dtype)

    rng = xp.random.default_rng(seed=seed)
    s = rng.standard_normal(size=(op.dim, (m + 2) // 4), dtype=dtype)
    g = rng.integers(0, 2, size=(op.dim, (m - 2) // 2)) * 2 - 1

    with pxrt.Precision(width):
        data = op.apply(s.T).T

    kwargs = dict(mode="reduced")
    if xp == pxd.NDArrayInfo.DASK.module():
        data = data.rechunk({0: "auto", 1: -1})
        kwargs.pop("mode", None)

    q, _ = xp.linalg.qr(data, **kwargs)
    proj = g - q @ (q.T @ g)

    tr = (op.apply(q.T) @ q).trace()
    tr += (2 / (m - 2)) * (op.apply(proj.T) @ proj).trace()
    return float(tr)



import collections.abc as cabc

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.opt.stop as pxst

# We do not test MaxIter(), ManualStop() and MaxDuration() since they are trivial.


class TestAbsError:
    @pytest.mark.parametrize(
        ["eps", "f", "state", "stop_val"],
        [
            [3, None, 4, False],
            [3, None, 3, True],
            [3, None, 2, True],
            [3, lambda _: _**2, 4, False],
            [3, lambda _: _**2, 3, False],
            [3, lambda _: _**2, np.sqrt(3), True],
            [3, lambda _: _**2, 1, True],
        ],
    )
    def test_scalar_in(self, eps, f, state, stop_val):
        sc = pxst.AbsError(eps=eps, f=f)
        state = dict(x=state)
        assert sc.stop(state) == stop_val
        sc.info()  # just to make sure it doesn't crash

    @pytest.mark.parametrize(
        ["eps", "f", "satisfy_all", "state", "stop_val"],
        [
            # 1 input, function
            [np.sqrt(14), None, True, np.arange(1, 4), True],
            [np.sqrt(14), None, True, np.arange(1, 5), False],
            [6, lambda _: _.sum(axis=-1, keepdims=True), True, np.broadcast_to(np.arange(1, 4), (1, 3)), True],
            [6, lambda _: _.sum(axis=-1, keepdims=True), True, np.broadcast_to(np.arange(1, 5), (1, 4)), False],
            # N input, satisfy_[any/all]
            [6.5, None, True, np.array([np.linspace(1, 4, 5), np.linspace(1, 5, 5)]), False],
            [6.5, None, False, np.array([np.linspace(1, 4, 5), np.linspace(1, 5, 5)]), True],
        ],
    )
    def test_array_in(self, eps, f, satisfy_all, state, stop_val, xp):
        sc = pxst.AbsError(eps=eps, f=f, satisfy_all=satisfy_all)
        state = dict(x=xp.array(state))  # test all possible array backends.
        assert sc.stop(state) == stop_val
        sc.info()  # just to make sure it doesn't crash


class TestRelError:
    @pytest.mark.parametrize(
        ["eps", "f", "state", "stop_val"],
        [
            [0.5, None, (1, 2), False],
            [0.5, None, (2, 3), True],
            [0.5, None, (3, 4), True],
            [0.3, lambda _: _**2, (1, 2), False],
            [0.3, lambda _: _**2, (2, 3), False],
            [0.9, lambda _: _**2, (3, 4), True],
        ],
    )
    def test_scalar_in(self, eps, f, state, stop_val):
        sc = pxst.RelError(eps=eps, f=f)
        state0 = dict(x=state[0])
        assert sc.stop(state0) is False
        state1 = dict(x=state[1])
        assert sc.stop(state1) == stop_val
        sc.info()  # just to make sure it doesn't crash

    @pytest.mark.parametrize(
        ["eps", "f", "satisfy_all", "state", "stop_val"],
        [
            # 1 input, function
            [0.6, None, True, (np.r_[1, 1, 1], np.r_[1, 2, 1]), True],
            [0.23, None, True, (np.r_[2, 2, 2], np.r_[2, 3, 2]), False],
            [
                0.4,
                lambda _: _.sum(axis=-1, keepdims=True),
                True,
                (
                    np.broadcast_to(np.r_[1, 1, 1], (1, 3)),
                    np.broadcast_to(np.r_[1, 2, 1], (1, 3)),
                ),
                True,
            ],
            [
                0.16,
                lambda _: _.sum(axis=-1, keepdims=True),
                True,
                (
                    np.broadcast_to(np.r_[2, 2, 2], (1, 3)),
                    np.broadcast_to(np.r_[2, 3, 2], (1, 3)),
                ),
                False,
            ],
            # N input, satisfy_[any/all]
            [
                0.3,
                None,
                True,
                (
                    np.array([[1, 1, 1], [2, 2, 2]]),
                    np.array([[1, 2, 1], [2, 3, 2]]),
                ),
                False,
            ],
            [
                0.3,
                None,
                False,
                (
                    np.array([[1, 1, 1], [2, 2, 2]]),
                    np.array([[1, 2, 1], [2, 3, 2]]),
                ),
                True,
            ],
        ],
    )
    def test_array_in(self, eps, f, satisfy_all, state, stop_val, xp):
        sc = pxst.RelError(eps=eps, f=f, satisfy_all=satisfy_all)
        state0 = dict(x=xp.array(state[0]))  # test all possible array backends.
        assert sc.stop(state0) is False
        state1 = dict(x=xp.array(state[1]))  # test all possible array backends.
        assert sc.stop(state1) == stop_val
        sc.info()  # just to make sure it doesn't crash


@pytest.mark.parametrize(
    ["sc", "state_stream"],
    [
        [pxst.MaxIter(n=10), [{}] * 12],  # state meaningless
        [pxst.ManualStop(), [{}] * 12],  # state meaningless
        # [pxst.MaxDuration(), [{}] * 10],  # MaxDuration is never 100% reproducible
        [pxst.AbsError(eps=3), [dict(x=_) for _ in np.arange(10, 0, -1)]],
        [pxst.RelError(eps=1 / 6), [dict(x=_) for _ in np.arange(10)]],
    ],
)
def test_clear(
    sc: pxa.StoppingCriterion,
    state_stream: cabc.Sequence[cabc.Mapping],
):
    """
    Verify that `StoppingCriterion.clear()` makes `sc` reusable.

    Algorithm:
    * feed `sc` with `state_stream`;
    * memoize stop()/info() outputs produced;
    * call clear();
    * verify that `sc` fed with `state_stream` again produces exactly the same output.
    """

    def memoize(sc, state_stream):
        h_stop, h_info = [], []  # history
        sc.clear()
        for state in state_stream:
            h_stop.append(sc.stop(state))
            h_info.append(sc.info())
        return h_stop, h_info

    h1_stop, h1_info = memoize(sc, state_stream)
    h2_stop, h2_info = memoize(sc, state_stream)
    assert h1_stop == h2_stop
    assert h1_info == h2_info



import math
import types
import typing as typ

import pyxu.info.ptype as pxt
import pyxu.util as pxu

__all__ = [
    "OnlineMoment",
    "OnlineCenteredMoment",
    "OnlineVariance",
    "OnlineStd",
    "OnlineSkewness",
    "OnlineKurtosis",
]


class _OnlineStat:
    """
    Abstract base class to compute online statistics based on outputs of a Sampler object.

    An _OnlineStat object should be paired with a single Sampler object.

    Composite _OnlineStat objects can be implemented via the overloaded +, -, *, and ** operators.
    """

    def __init__(self):
        self._num_samples = 0
        self._stat = None

    def update(self, x: pxt.NDArray) -> pxt.NDArray:
        """
        Update the online statistic based on a new sample. Should update `_num_samples` and `_stat` attributes.

        Parameters
        ----------
        x: NDArray
            New sample.

        Returns
        -------
        stat: NDArray
            Updated statistic.
        """
        raise NotImplementedError

    def stat(self) -> pxt.NDArray:
        """Get current online statistic estimate."""
        return self._stat

    def __add__(self, other: "_OnlineStat"):
        stat = _OnlineStat()
        stat_update = lambda _, x: self.update(x) + other.update(x)
        stat.update = types.MethodType(stat_update, stat)
        return stat

    def __sub__(self, other: "_OnlineStat"):
        stat = _OnlineStat()
        stat_update = lambda _, x: self.update(x) - other.update(x)
        stat.update = types.MethodType(stat_update, stat)
        return stat

    def __mul__(self, other: typ.Union[pxt.Real, pxt.Integer, "_OnlineStat"]):
        stat = _OnlineStat()
        if isinstance(other, _OnlineStat):
            stat_update = lambda _, x: self.update(x) * other.update(x)
        elif isinstance(other, pxt.Real) or isinstance(other, pxt.Integer):
            stat_update = lambda _, x: self.update(x) * other
        else:
            return NotImplemented
        stat.update = types.MethodType(stat_update, stat)
        return stat

    def __rmul__(self, other: typ.Union[pxt.Real, pxt.Integer]):
        return self.__mul__(other)

    def __truediv__(self, other: typ.Union[pxt.Real, pxt.Integer, "_OnlineStat"]):
        stat = _OnlineStat()
        if isinstance(other, _OnlineStat):

            def stat_update(stat, x):
                xp = pxu.get_array_module(x)
                out = xp.divide(self.update(x), other.update(x), where=(other.update(x) != 0))
                out[other.update(x) == 0] = xp.nan
                return out

        elif isinstance(other, pxt.Real) or isinstance(other, pxt.Integer):
            stat_update = lambda _, x: self.update(x) / other
        else:
            return NotImplemented
        stat.update = types.MethodType(stat_update, stat)
        return stat

    def __pow__(self, expo: typ.Union[pxt.Real, pxt.Integer]):
        if not (isinstance(expo, pxt.Real) or isinstance(expo, pxt.Integer)):
            return NotImplemented
        stat = _OnlineStat()
        stat_update = lambda _, x: self.update(x) ** expo
        stat.update = types.MethodType(stat_update, stat)
        return stat


class OnlineMoment(_OnlineStat):
    r"""
    Pointwise online moment.

    For :math:`d \geq 1`, the :math:`d`-th order centered moment of the :math:`K` samples :math:`(\mathbf{x}_k)_{1 \leq
    k \leq K}` is given by :math:`\frac{1}{K}\sum_{k=1}^K \mathbf{x}_k^d`.
    """

    def __init__(self, order: pxt.Real = 1):
        super().__init__()
        self._order = order

    def update(self, x: pxt.NDArray) -> pxt.NDArray:
        if self._num_samples == 0:
            self._stat = x**self._order
        else:
            self._stat *= self._num_samples
            self._stat += x**self._order
        self._num_samples += 1
        self._stat /= self._num_samples
        return self._stat


class OnlineCenteredMoment(_OnlineStat):
    r"""
    Pointwise online centered moment.

    For :math:`d \geq 2`, the :math:`d`-th order centered moment of the :math:`K` samples :math:`(\mathbf{x}_k)_{1 \leq
    k \leq K}` is given by :math:`\boldsymbol{\mu}_d=\frac{1}{K} \sum_{k=1}^K (\mathbf{x}_k-\boldsymbol{\mu})^d`, where
    :math:`\boldsymbol{\mu}` is the sample mean.

    Notes
    -----
    This class implements the *Welford algorithm* described in [WelfordAlg]_, which is a numerically stable algorithm
    for computing online centered moments. In particular, it avoids `catastrophic cancellation
    <https://en.wikipedia.org/wiki/Catastrophic_cancellation>`_ issues that may arise when naively computing online
    centered moments, which would lead to a loss of numerical precision.

    Note that this class internally stores the values of all online centered moments of order :math:`d'` for :math:`2
    \leq d' \leq d` in the attribute ``_corrected_sums`` as well as the online mean (``_mean`` attribute). More
    precisely, the array ``_corrected_sums[i, :]`` corresponds to the online sum :math:`\boldsymbol{\mu}_{i+2}=
    \sum_{k=1}^K (\mathbf{x}_k-\boldsymbol{\mu})^{i+2}` for :math:`0 \leq i \leq d-2`.
    """

    def __init__(self, order: pxt.Real = 2):
        super().__init__()
        self._order = order
        self._corrected_sums = None
        self._mean = None

    def update(self, x: pxt.NDArray) -> pxt.NDArray:
        if self._num_samples == 0:
            xp = pxu.get_array_module(x)
            self._corrected_sums = xp.zeros((self._order - 1,) + x.shape)
            self._mean = x.copy()
        else:
            temp = (x - self._mean) / (self._num_samples + 1)
            for r in range(self._order, 1, -1):  # Update in descending order because updates depend on lower orders
                for s in range(2, r):  # s = r term excluded since it corresponds to previous iterate
                    self._corrected_sums[r - 2, :] += (
                        math.comb(r, s) * self._corrected_sums[s - 2, :] * (-temp) ** (r - s)
                    )
                self._corrected_sums[r - 2, :] += self._num_samples * (-temp) ** r  # Contribution of s = 0 term
                self._corrected_sums[r - 2, :] += (self._num_samples * temp) ** r
            self._mean *= self._num_samples / (self._num_samples + 1)
            self._mean += x / (self._num_samples + 1)
        self._num_samples += 1
        self._stat = self._corrected_sums[-1, :] / self._num_samples
        return self._stat


def OnlineVariance():
    r"""
    Pointwise online variance.

    The pointwise online variance of the :math:`K` samples :math:`(\mathbf{x}_k)_{1 \leq k \leq K}` is given by
    :math:`\boldsymbol{\sigma}^2 = \frac{1}{K}\sum_{k=1}^K (\mathbf{x}_k-\boldsymbol{\mu})^2`, where
    :math:`\boldsymbol{\mu}` is the sample mean.
    """
    return OnlineCenteredMoment(order=2)


def OnlineStd():
    r"""
    Pointwise online standard deviation.

    The pointwise online standard deviation of the :math:`K` samples :math:`(\mathbf{x}_k)_{1 \leq k \leq K}` is given
    by :math:`\boldsymbol{\sigma}=\sqrt{\frac{1}{K}\sum_{k=1}^K (\mathbf{x}_k - \boldsymbol{\mu})^2}`, where
    :math:`\boldsymbol{\mu}` is the sample mean.
    """
    return OnlineVariance() ** (1 / 2)


def OnlineSkewness():
    r"""
    Pointwise online skewness.

    The pointwise online skewness of the :math:`K` samples :math:`(\mathbf{x}_k)_{1 \leq k \leq K}` is given by
    :math:`\frac{1}{K}\sum_{k=1}^K \left( \frac{\mathbf{x}_k-\boldsymbol{\mu}}{\boldsymbol{\sigma}}\right)^3`, where
    :math:`\boldsymbol{\mu}` is the sample mean and :math:`\boldsymbol{\sigma}` its standard deviation.

    `Skewness <https://en.wikipedia.org/wiki/Skewness>`_ is a measure of asymmetry of a distribution around its mean.
    Negative skewness indicates that the distribution has a heavier tail on the left side than on the right side,
    positive skewness indicates the opposite, and values close to zero indicate a symmetric distribution.
    """
    return OnlineCenteredMoment(order=3) / OnlineStd() ** 3


def OnlineKurtosis():
    r"""
    Pointwise online kurtosis.

    The pointwise online variance of the :math:`K` samples :math:`(\mathbf{x}_k)_{1 \leq k \leq K}` is given by
    :math:`\frac{1}{K}\sum_{k=1}^K \left( \frac{\mathbf{x}_k-\boldsymbol{\mu}}{\boldsymbol{\sigma}}\right)^4`, where
    :math:`\boldsymbol{\mu}` is the sample mean and :math:`\boldsymbol{\sigma}` its standard deviation.

    `Kurtosis <https://en.wikipedia.org/wiki/Kurtosis>`_ is a measure of the heavy-tailedness of a distribution. In
    particular, the kurtosis of a Gaussian distribution is always 3.
    """
    return OnlineCenteredMoment(order=4) / OnlineStd() ** 4



import numpy as np

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.math as pxm
import pyxu.runtime as pxrt

__all__ = [
    "NLCG",
]


class NLCG(pxa.Solver):
    r"""
    Nonlinear Conjugate Gradient Method (NLCG).

    The Nonlinear Conjugate Gradient method finds a local minimum of the problem

    .. math::

       \min_{x\in\mathbb{R}^{N}} f(x),

    where :math:`f: \mathbb{R}^{N} \to \mathbb{R}` is a *differentiable* functional.  When :math:`f` is quadratic, NLCG
    is equivalent to the Conjugate Gradient (CG) method.  NLCG hence has similar convergence behaviour to CG if
    :math:`f` is locally-quadratic.  The converge speed may be slower however due to its line-search overhead
    [NumOpt_NocWri]_.

    The norm of the `gradient <https://www.wikiwand.com/en/Nonlinear_conjugate_gradient_method>`_ :math:`\nabla f_k =
    \nabla f(x_k)` is used as the default stopping criterion.  By default, the iterations stop when the norm of the
    gradient is smaller than 1e-4.

    Multiple variants of NLCG exist.  They differ mainly in how the weights applied to conjugate directions are updated.
    Two popular variants are implemented:

    * The Fletcher-Reeves variant:

      .. math::

         \beta_k^\text{FR}
         =
         \frac{
           \Vert{\nabla f_{k+1}}\Vert_{2}^{2}
         }{
           \Vert{\nabla f_{k}}\Vert_{2}^{2}
         }.

    * The Polak-Ribière+ method:

      .. math::

         \beta_k^\text{PR}
         =
         \frac{
           \nabla f_{k+1}^T\left(\nabla f_{k+1} - \nabla f_k\right)
         }{
           \Vert{\nabla f_{k}}\Vert_{2}^{2}
         } \\
         \beta_k^\text{PR+}
         =
         \max\left(0, \beta_k^\text{PR}\right).

    Parameters (``__init__()``)
    ---------------------------
    * **f** (:py:class:`~pyxu.abc.DiffFunc`)
      --
      Differentiable function :math:`\mathcal{F}`.
    * **\*\*kwargs** (:py:class:`~collections.abc.Mapping`)
      --
      Other keyword parameters passed on to :py:meth:`pyxu.abc.Solver.__init__`.

    Parameters (``fit()``)
    ----------------------
    * **x0** (:py:attr:`~pyxu.info.ptype.NDArray`)
      --
      (..., N) initial point(s).
    * **variant** ("PR", "FR")
      --
      Name of the NLCG variant to use:

      * "PR": Polak-Ribière+ variant (default).
      * "FR": Fletcher-Reeves variant.
    * **restart_rate** (:py:attr:`~pyxu.info.ptype.Integer`, :py:obj:`None`)
      --
      Number of iterations after which restart is applied.

      By default, restart is done after :math:`N` iterations.
    * **\*\*kwargs** (:py:class:`~collections.abc.Mapping`)
      --
      Optional parameters forwarded to :py:func:`~pyxu.math.backtracking_linesearch`.

      If `a0` is unspecified and :math:`\nabla f` is :math:`\beta`-Lipschitz continuous, then `a0` is auto-chosen as
      :math:`\beta^{-1}`.  Users are expected to set `a0` if its value cannot be auto-inferred.

      Other keyword parameters are passed on to :py:meth:`pyxu.abc.Solver.fit`.

    Example
    -------
    Consider the following quadratic optimization problem:

    .. math:

       \min_{\mathbf{x}} \Vert{A\mathbf{x}-\mathbf{b}}\Vert_2^2


    This problem is strictly convex, hence NLCG will converge to the optimal solution:

    .. code-block:: python3

       import numpy as np

       import pyxu.operator as pxo
       import pyxu.opt.solver as pxsl

       N, a, b = 5, 3, 1
       f = pxo.SquaredL2Norm(N).asloss(b).argscale(a)  # \norm(Ax - b)**2

       nlcg = pxsl.NLCG(f)
       nlcg.fit(x0=np.zeros((N,)), variant="FR")
       x_opt = nlcg.solution()
       np.allclose(x_opt, 1/a)  # True

    Note however that the CG method is preferable in this context since it omits the linesearch overhead. The former
    depends on the cost of applying :math:`A`, and may be significant.
    """

    def __init__(self, f: pxa.DiffFunc, **kwargs):
        kwargs.update(
            log_var=kwargs.get("log_var", ("x",)),
        )
        super().__init__(**kwargs)

        self._f = f

    @pxrt.enforce_precision(i="x0")
    def m_init(
        self,
        x0: pxt.NDArray,
        variant: str = "PR",
        restart_rate: pxt.Integer = None,
        **kwargs,
    ):
        mst = self._mstate  # shorthand

        if (a0 := kwargs.get("a0")) is None:
            d_l = self._f.diff_lipschitz
            if np.isclose(d_l, np.inf) or np.isclose(d_l, 0):
                msg = "[NLCG] cannot auto-infer initial step size: specify `a0` manually in NLCG.fit()"
                raise ValueError(msg)
            else:
                a0 = pxrt.coerce(1 / d_l)

        if restart_rate is not None:
            assert restart_rate >= 1
            mst["restart_rate"] = int(restart_rate)
        else:
            mst["restart_rate"] = x0.shape[-1]

        import pyxu.math.linesearch as ls

        mst["x"] = x0
        mst["gradient"] = self._f.grad(x0)
        mst["conjugate_dir"] = -mst["gradient"].copy()
        mst["variant"] = self.__parse_variant(variant)
        mst["ls_a0"] = a0
        mst["ls_r"] = kwargs.get("r", ls.LINESEARCH_DEFAULT_R)
        mst["ls_c"] = kwargs.get("c", ls.LINESEARCH_DEFAULT_C)
        mst["ls_a_k"] = mst["ls_a0"]

    def m_step(self):
        mst = self._mstate  # shorthand
        x_k, g_k, p_k = mst["x"], mst["gradient"], mst["conjugate_dir"]

        a_k = pxm.backtracking_linesearch(
            f=self._f,
            x=x_k,
            gradient=g_k,
            direction=p_k,
            a0=mst["ls_a0"],
            r=mst["ls_r"],
            c=mst["ls_c"],
        )
        # In-place implementation of -----------------
        #   x_kp1 = x_k + p_k * a_k
        x_kp1 = p_k.copy()
        x_kp1 *= a_k
        x_kp1 += x_k
        # --------------------------------------------
        g_kp1 = self._f.grad(x_kp1)

        # Because NLCG can only generate n conjugate vectors in an n-dimensional space, it makes sense
        # to restart NLCG every n iterations.
        if self._astate["idx"] % mst["restart_rate"] == 0:
            beta_kp1 = pxrt.coerce(0)
        else:
            beta_kp1 = self.__compute_beta(g_k, g_kp1)

        # In-place implementation of -----------------
        #   p_kp1 = -g_kp1 + beta_kp1 * p_k
        p_kp1 = p_k.copy()
        p_kp1 *= beta_kp1
        p_kp1 -= g_kp1
        # --------------------------------------------

        mst["x"], mst["gradient"], mst["conjugate_dir"], mst["ls_a_k"] = x_kp1, g_kp1, p_kp1, a_k

    def default_stop_crit(self) -> pxa.StoppingCriterion:
        from pyxu.opt.stop import AbsError

        stop_crit = AbsError(
            eps=1e-4,
            var="gradient",
            f=None,
            norm=2,
            satisfy_all=True,
        )
        return stop_crit

    def objective_func(self) -> pxt.NDArray:
        return self._f(self._mstate["x"])

    def solution(self) -> pxt.NDArray:
        """
        Returns
        -------
        x: NDArray
            (..., N) solution.
        """
        data, _ = self.stats()
        return data.get("x")

    def __compute_beta(self, g_k: pxt.NDArray, g_kp1: pxt.NDArray) -> pxt.NDArray:
        v = self._mstate["variant"]
        if v == "fr":  # Fletcher-Reeves
            gn_k = pxm.norm(g_k, axis=-1, keepdims=True)
            gn_kp1 = pxm.norm(g_kp1, axis=-1, keepdims=True)
            beta = (gn_kp1 / gn_k) ** 2
        elif v == "pr":  # Poliak-Ribière+
            gn_k = pxm.norm(g_k, axis=-1, keepdims=True)
            numerator = (g_kp1 * (g_kp1 - g_k)).sum(axis=-1, keepdims=True)
            beta = numerator / (gn_k**2)
            beta = beta.clip(min=0)
        return beta  # (..., 1)

    def __parse_variant(self, variant: str) -> str:
        supported_variants = {"fr", "pr"}
        if (v := variant.lower().strip()) not in supported_variants:
            raise ValueError(f"Unsupported variant '{variant}'.")
        return v



import itertools
import math
import warnings

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.info.warning as pxw
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "PGD",
]


class PGD(pxa.Solver):
    r"""
    Proximal Gradient Descent (PGD) solver.

    PGD solves minimization problems of the form

    .. math::

       {\min_{\mathbf{x}\in\mathbb{R}^N} \;\mathcal{F}(\mathbf{x})\;\;+\;\;\mathcal{G}(\mathbf{x})},

    where:

    * :math:`\mathcal{F}:\mathbb{R}^N\rightarrow \mathbb{R}` is *convex* and *differentiable*, with
      :math:`\beta`-*Lipschitz continuous* gradient, for some :math:`\beta\in[0,+\infty[`.
    * :math:`\mathcal{G}:\mathbb{R}^N\rightarrow \mathbb{R}\cup\{+\infty\}` is a *proper*, *lower semicontinuous* and
      *convex function* with a *simple proximal operator*.

    Remarks
    -------
    * The problem is *feasible* -- i.e. there exists at least one solution.

    * The algorithm is still valid if either :math:`\mathcal{F}` or :math:`\mathcal{G}` is zero.

    * The convergence is guaranteed for step sizes :math:`\tau\leq 1/\beta`.

    * Various acceleration schemes are described in [APGD]_.  PGD achieves the following (optimal) *convergence rate*
      with the implemented acceleration scheme from Chambolle & Dossal:

      .. math::

         \lim\limits_{n\rightarrow \infty} n^2\left\vert \mathcal{J}(\mathbf{x}^\star)- \mathcal{J}(\mathbf{x}_n)\right\vert=0
         \qquad\&\qquad
         \lim\limits_{n\rightarrow \infty} n^2\Vert \mathbf{x}_n-\mathbf{x}_{n-1}\Vert^2_\mathcal{X}=0,

      for *some minimiser* :math:`{\mathbf{x}^\star}\in\arg\min_{\mathbf{x}\in\mathbb{R}^N}
      \;\left\{\mathcal{J}(\mathbf{x}):=\mathcal{F}(\mathbf{x})+\mathcal{G}(\mathbf{x})\right\}`.  In other words, both
      the objective functional and the PGD iterates :math:`\{\mathbf{x}_n\}_{n\in\mathbb{N}}` converge at a rate
      :math:`o(1/n^2)`.  Significant practical *speedup* can be achieved for values of :math:`d` in the range
      :math:`[50,100]` [APGD]_.

    * The relative norm change of the primal variable is used as the default stopping criterion.  By default, the
      algorithm stops when the norm of the difference between two consecutive PGD iterates
      :math:`\{\mathbf{x}_n\}_{n\in\mathbb{N}}` is smaller than 1e-4.  Different stopping criteria can be used.

    Parameters (``__init__()``)
    ---------------------------
    * **f** (:py:class:`~pyxu.abc.DiffFunc`, :py:obj:`None`)
      --
      Differentiable function :math:`\mathcal{F}`.
    * **g** (:py:class:`~pyxu.abc.ProxFunc`, :py:obj:`None`)
      --
      Proximable function :math:`\mathcal{G}`.
    * **\*\*kwargs** (:py:class:`~collections.abc.Mapping`)
      --
      Other keyword parameters passed on to :py:meth:`pyxu.abc.Solver.__init__`.

    Parameters (``fit()``)
    ----------------------
    * **x0** (:py:attr:`~pyxu.info.ptype.NDArray`)
      --
      (..., N) initial point(s).
    * **tau** (:py:attr:`~pyxu.info.ptype.Real`, :py:obj:`None`)
      --
      Gradient step size.  Defaults to :math:`1 / \beta` if unspecified.
    * **acceleration** (:py:obj:`bool`)
      --
      If True (default), then use Chambolle & Dossal acceleration scheme.
    * **d** (:py:attr:`~pyxu.info.ptype.Real`)
      --
      Chambolle & Dossal acceleration parameter :math:`d`.  Should be greater than 2.  Only meaningful if `acceleration`
      is True.  Defaults to 75 in unspecified.
    * **\*\*kwargs** (:py:class:`~collections.abc.Mapping`)
      --
      Other keyword parameters passed on to :py:meth:`pyxu.abc.Solver.fit`.
    """

    def __init__(
        self,
        f: pxa.DiffFunc = None,
        g: pxa.ProxFunc = None,
        **kwargs,
    ):
        kwargs.update(
            log_var=kwargs.get("log_var", ("x",)),
        )
        super().__init__(**kwargs)

        if (f is None) and (g is None):
            msg = " ".join(
                [
                    "Cannot minimize always-0 functional.",
                    "At least one of Parameter[f, g] must be specified.",
                ]
            )
            raise ValueError(msg)
        else:
            # Problem
            # -------
            # If f/g is domain-agnostic and g/f is unspecified, cannot auto-infer NullFunc
            # dimension.
            #
            # Solution
            # --------
            # Delay initialization of missing f/g to m_init(), where x0's shape can be used.
            self._f = f
            self._g = g

    @pxrt.enforce_precision(i=("x0", "tau"))
    def m_init(
        self,
        x0: pxt.NDArray,
        tau: pxt.Real = None,
        acceleration: bool = True,
        d: pxt.Real = 75,
    ):
        mst = self._mstate  # shorthand
        mst["x"] = mst["x_prev"] = x0

        if self._f is None:
            self._f = pxo.NullFunc(dim=x0.shape[-1])
        if self._g is None:
            self._g = pxo.NullFunc(dim=x0.shape[-1])

        if tau is None:
            mst["tau"] = pxrt.coerce(1 / self._f.diff_lipschitz)
            if math.isinf(mst["tau"]):
                # _f is constant-valued: \tau is a free parameter.
                mst["tau"] = 1
                msg = "\n".join(
                    [
                        rf"The gradient/proximal step size \tau is auto-set to {mst['tau']}.",
                        r"Choosing \tau manually may lead to faster convergence.",
                    ]
                )
                warnings.warn(msg, pxw.AutoInferenceWarning)
        else:
            try:
                assert tau > 0
                mst["tau"] = tau
            except Exception:
                raise ValueError(f"tau must be positive, got {tau}.")

        if acceleration:
            try:
                assert d > 2
                mst["a"] = (pxrt.coerce(k / (k + 1 + d)) for k in itertools.count(start=0))
            except Exception:
                raise ValueError(f"Expected d > 2, got {d}.")
        else:
            mst["a"] = itertools.repeat(pxrt.coerce(0))

    def m_step(self):
        mst = self._mstate  # shorthand
        a = next(mst["a"])

        # In-place implementation of -----------------
        #   y = (1 + a) * mst["x"] - a * mst["x_prev"]
        y = mst["x"] - mst["x_prev"]
        y *= a
        y += mst["x"]
        # --------------------------------------------

        # In-place implementation of -----------------
        #   z = y - mst["tau"] * self._f.grad(y)
        z = pxu.copy_if_unsafe(self._f.grad(y))
        z *= -mst["tau"]
        z += y
        # --------------------------------------------

        mst["x_prev"], mst["x"] = mst["x"], self._g.prox(z, mst["tau"])

    def default_stop_crit(self) -> pxa.StoppingCriterion:
        from pyxu.opt.stop import RelError

        stop_crit = RelError(
            eps=1e-4,
            var="x",
            f=None,
            norm=2,
            satisfy_all=True,
        )
        return stop_crit

    def objective_func(self) -> pxt.NDArray:
        func = lambda x: self._f.apply(x) + self._g.apply(x)

        y = func(self._mstate["x"])
        return y

    def solution(self) -> pxt.NDArray:
        """
        Returns
        -------
        x: NDArray
            (..., N) solution.
        """
        data, _ = self.stats()
        return data.get("x")



# How coo_block() tests work:
#
# * COOBlockMixin auto-defines all (input,output) pairs.
#   [Caveat: we assume the base operators (op_blocks) are correctly implemented.]
#   (True if choosing test operators from examples/.)
#   [Caveat: we assume the punch-out indices (op_hole) are not out-of-bound.]
#
# * To test a coo_block-ed operator, inherit from COOBlockMixin and the suitable conftest.MapT
#   subclass which the compound operator should abide by.

import collections.abc as cabc
import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


# test operators --------------------------------------------------------------
def op_map(dim: int):
    import pyxu_tests.operator.examples.test_map as tc

    return tc.ReLU(M=dim)


def op_diffmap(dim: int):
    import pyxu_tests.operator.examples.test_diffmap as tc

    return tc.Sin(M=dim)


def op_linop(dim: int, codim_scale: int):
    import pyxu_tests.operator.examples.test_linop as tc

    return tc.Tile(N=dim, M=codim_scale)


def op_squareop(dim: int):
    import pyxu_tests.operator.examples.test_squareop as tc

    return tc.CumSum(N=dim)


def op_normalop(dim: int):
    import pyxu_tests.operator.examples.test_normalop as tc

    rng = np.random.default_rng(seed=2)
    h = rng.normal(size=(dim,))
    return tc.CircularConvolution(h=h)


def op_unitop(dim: int):
    import pyxu_tests.operator.examples.test_unitop as tc

    return tc.Permutation(N=dim)


def op_selfadjointop(dim: int):
    import pyxu_tests.operator.examples.test_selfadjointop as tc

    return tc.SelfAdjointConvolution(N=dim)


def op_posdefop(dim: int):
    import pyxu_tests.operator.examples.test_posdefop as tc

    return tc.PSDConvolution(N=dim)


def op_projop(dim: int):
    import pyxu_tests.operator.examples.test_projop as tc

    return tc.Oblique(N=dim, alpha=np.pi / 4)


def op_orthprojop(dim: int):
    import pyxu_tests.operator.examples.test_orthprojop as tc

    return tc.ScaleDown(N=dim)


def op_difffunc(dim: int):
    import pyxu_tests.operator.examples.test_difffunc as tc

    return tc.SquaredL2Norm(M=dim)


def op_proxfunc(dim: int):
    import pyxu_tests.operator.examples.test_proxfunc as tc

    return tc.L1Norm(M=dim)


def op_proxdifffunc(dim: int):
    import pyxu_tests.operator.examples.test_proxdifffunc as tc

    return tc.SquaredL2Norm(M=dim)


def op_quadraticfunc(dim: int):
    from pyxu_tests.operator.examples.test_linfunc import ScaledSum
    from pyxu_tests.operator.examples.test_posdefop import PSDConvolution

    return pxa.QuadraticFunc(
        shape=(1, dim),
        Q=PSDConvolution(N=dim),
        c=ScaledSum(N=dim),
        t=1,
    )


def op_linfunc(dim: int):
    import pyxu_tests.operator.examples.test_linfunc as tc

    op = tc.ScaledSum(N=dim)
    return op


# Data Mixin ------------------------------------------------------------------
class COOBlockMixin:
    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def op_blocks(self) -> list[list[pxt.OpT]]:
        # Override in inherited class with a 2D nested sequence of operators to be fed to block(, order=1).
        raise NotImplementedError

    @pytest.fixture
    def op_hole(self) -> list[tuple[int, int]]:
        # Override in inherited class with grid indices to punch out of `op_blocks()`.
        raise NotImplementedError

    @pytest.fixture
    def op_gt(self, op_blocks, op_hole) -> pxt.OpT:
        # Replace holes in op_block with 0s -> will serve as the ground-truth.
        op_GT = []
        for _i, row in enumerate(op_blocks):
            op_GT.append([])
            for _j, op in enumerate(row):
                if (_i, _j) in op_hole:
                    _op = pxo.NullOp(shape=op.shape)
                else:
                    _op = op
                op_GT[_i].append(_op)

        op = pxo.block(op_GT, order=1)
        return op

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
            [True, False],  # parallel
        )
    )
    def spec(self, op_blocks, op_hole, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width, parallel = request.param

        data, i, j = [], [], []
        for _i, row in enumerate(op_blocks):
            for _j, op in enumerate(row):
                if (_i, _j) not in op_hole:
                    data.append(op)
                    i.append(_i)
                    j.append(_j)
        N_row, N_col = max(i) + 1, max(j) + 1
        op = pxo.coo_block(
            (data, (i, j)),
            grid_shape=(N_row, N_col),
            parallel=parallel,
        )
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, op_gt) -> pxt.OpShape:
        return op_gt.shape

    @pytest.fixture
    def data_apply(self, op_gt) -> conftest.DataLike:
        arr = self._random_array((op_gt.dim,), seed=17)  # random seed for reproducibility
        out = op_gt.apply(arr)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_adjoint(self, op_gt) -> conftest.DataLike:
        arr = self._random_array((op_gt.codim,), seed=17)  # random seed for reproducibility
        out = op_gt.adjoint(arr)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_grad(self, op_gt) -> conftest.DataLike:
        arr = self._random_array((op_gt.dim,), seed=17)  # random seed for reproducibility
        out = op_gt.grad(arr)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_prox(self, op_gt) -> conftest.DataLike:
        arr = self._random_array((op_gt.dim,), seed=17)  # random seed for reproducibility
        tau = np.abs(self._random_array((1,), seed=31))[0]  # random seed for reproducibility
        out = op_gt.prox(arr=arr, tau=tau)
        return dict(
            in_=dict(
                arr=arr,
                tau=tau,
            ),
            out=out,
        )

    @pytest.fixture
    def data_math_lipschitz(self, data_shape) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, data_shape[1]
        return self._random_array((N_test, dim))

    @pytest.fixture
    def data_math_diff_lipschitz(self, data_shape) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, data_shape[1]
        return self._random_array((N_test, dim))


class COOBlockFuncMixin(COOBlockMixin):
    @pytest.fixture
    def op_hole(self) -> list[tuple[int, int]]:
        return []  # horizontal coo_block() objects cannot have any holes.

    # Tests -------------------------------------------------------------------
    def test_interface_asloss(self, op, xp, width):
        self._skip_if_disabled()
        with pytest.raises(NotImplementedError):
            # asloss() is ambiguous for block-defined operators -> must fail.
            op.asloss()


# Test classes (Maps) ---------------------------------------------------------
class TestCOOBlockMap(COOBlockMixin, conftest.MapT):
    @pytest.fixture(
        params=[
            [
                [op_map(3), op_diffmap(3)],
                [op_proxdifffunc(3), op_proxfunc(3)],
            ]
        ]
    )
    def op_blocks(self, request):
        return request.param

    @pytest.fixture(
        params=[
            [(0, 1), (1, 0)],
        ]
    )
    def op_hole(self, request):
        return request.param


class TestCOOBlockDiffMap(COOBlockMixin, conftest.DiffMapT):
    @pytest.fixture(
        params=[
            [
                [op_linfunc(3), op_difffunc(3)],
                [op_squareop(3), op_diffmap(3)],
            ]
        ]
    )
    def op_blocks(self, request):
        return request.param

    @pytest.fixture(
        params=[
            [(0, 0), (1, 1)],
        ]
    )
    def op_hole(self, request):
        return request.param


class TestCOOBlockLinOp(COOBlockMixin, conftest.LinOpT):
    @pytest.fixture(
        params=[
            [
                [op_linop(3, 2), op_normalop(6)],
                [op_linfunc(3), op_linfunc(6)],
            ]
        ]
    )
    def op_blocks(self, request):
        return request.param

    @pytest.fixture(
        params=[
            [(0, 1), (1, 0)],
        ]
    )
    def op_hole(self, request):
        return request.param


class TestCOOBlockSquareOp(COOBlockMixin, conftest.SquareOpT):
    @pytest.fixture(
        params=[
            [
                [op_squareop(7), op_posdefop(7)],
                [op_orthprojop(7), op_unitop(7)],
            ]
        ]
    )
    def op_blocks(self, request):
        return request.param

    @pytest.fixture(
        params=[
            [(0, 1), (1, 0)],
        ]
    )
    def op_hole(self, request):
        return request.param


# Test classes (Funcs) --------------------------------------------------------
class TestCOOBlockFunc(COOBlockFuncMixin, conftest.FuncT):
    @pytest.fixture(params=[])
    def op_blocks(self, request):
        return request.param


class TestCOOBlockProxFunc(COOBlockFuncMixin, conftest.ProxFuncT):
    @pytest.fixture(
        params=[
            [[op_proxfunc(7), op_proxdifffunc(4)]],
        ]
    )
    def op_blocks(self, request):
        return request.param


class TestCOOBlockDiffFunc(COOBlockFuncMixin, conftest.DiffFuncT):
    @pytest.fixture(
        params=[
            [[op_difffunc(5), op_proxdifffunc(7)]],
            [[op_difffunc(7), op_linfunc(8)]],
        ]
    )
    def op_blocks(self, request):
        return request.param


class TestCOOBlockProxDiffFunc(COOBlockFuncMixin, conftest.ProxDiffFuncT):
    @pytest.fixture(
        params=[
            [[op_proxdifffunc(3), op_quadraticfunc(5)]],
            [[op_proxdifffunc(3), op_linfunc(5)]],
        ]
    )
    def op_blocks(self, request):
        return request.param


class TestCOOBlockQuadraticFunc(COOBlockFuncMixin, conftest.QuadraticFuncT):
    @pytest.fixture(
        params=[
            [[op_quadraticfunc(5), op_linfunc(3)]],
            [[op_quadraticfunc(5), op_quadraticfunc(7)]],
        ]
    )
    def op_blocks(self, request):
        return request.param


class TestCOOBlockLinFunc(COOBlockFuncMixin, conftest.LinFuncT):
    @pytest.fixture(
        params=[
            [[op_linfunc(3), op_linfunc(1)]],
        ]
    )
    def op_blocks(self, request):
        return request.param



import collections.abc as cabc
import types
import typing as typ

import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "from_source",
]


def from_source(
    cls: pxt.OpC,
    shape: pxt.OpShape,
    embed: dict = None,
    vectorize: pxt.VarName = frozenset(),
    vmethod: typ.Union[str, dict] = None,
    enforce_precision: pxt.VarName = frozenset(),
    **kwargs,
) -> pxt.OpT:
    r"""
    Define an :py:class:`~pyxu.abc.Operator` from low-level constructs.

    Parameters
    ----------
    cls: OpC
        Operator sub-class to instantiate.
    shape: OpShape
        (N, M) operator shape.
    embed: dict
        (k[str], v[value]) pairs to embed into the created operator.

        `embed` is useful to attach extra information to synthesized :py:class:`~pyxu.abc.Operator` used by arithmetic
        methods.
    kwargs: dict
        (k[str], v[callable]) pairs to use as arithmetic methods.

        Keys must be entries from :py:meth:`~pyxu.abc.Property.arithmetic_methods`.

        Omitted arithmetic attributes/methods default to those provided by `cls`.
    vectorize: VarName
        Arithmetic methods to vectorize.

        `vectorize` is useful if an arithmetic method provided to `kwargs` (ex: :py:meth:`~pyxu.abc.Map.apply`):

        * does not support stacking dimensions; OR
        * does not support DASK inputs.
    vmethod: str, dict
        Vectorization strategy. (See :py:class:`~pyxu.util.vectorize`.)

        Different strategies can be applied per arithmetic method via a dictionary.
    enforce_precision: VarName
        Arithmetic methods to make compliant with Pyxu's runtime FP-precision.

        `enforce_precision` is useful if an arithmetic method provided to `kwargs` (ex: :py:meth:`~pyxu.abc.Map.apply`)
        is unaware of Pyxu's runtime FP context.

    Returns
    -------
    op: OpT
        (N, M) Pyxu-compliant operator.

    Notes
    -----
    * If provided, arithmetic methods must abide exactly to the Pyxu interface.  In particular, the following arithmetic
      methods, if supplied, **must** have the following interface:

      .. code-block:: python3

         def apply(self, arr: pxt.NDArray) -> pxt.NDArray
         def grad(self, arr: pxt.NDArray) -> pxt.NDArray
         def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray
         def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray
         def pinv(self, arr: pxt.NDArray, damp: pxt.Real) -> pxt.NDArray

      Moreover, the methods above **must** accept ``(..., M)``-shaped inputs for ``arr``.  If this does not hold,
      consider populating `vectorize`.

    * Auto-vectorization consists in decorating `kwargs`-specified arithmetic methods with
      :py:func:`~pyxu.util.vectorize`.  Auto-vectorization may be less efficient than explicitly providing a vectorized
      implementation.

    * Enforcing precision consists in decorating `kwargs`-specified arithmetic methods with
      :py:func:`~pyxu.runtime.enforce_precision`.  Not all arithmetic methods can be made runtime FP-precision
      compliant.  It is thus recommended to make arithmetic methods precision-compliant manually.

    Examples
    --------
    Creation of the custom element-wise differential operator :math:`f(\mathbf{x}) = \mathbf{x}^{2}`.

    .. code-block:: python3

       N = 5
       f = from_source(
           cls=pyxu.abc.DiffMap,
           shape=(N, N),
           apply=lambda _, arr: arr**2,
       )
       x = np.arange(N)
       y = f(x)  # [0, 1, 4, 9, 16]
       dL = f.diff_lipschitz  # inf (default value provided by DiffMap class.)

    In practice we know that :math:`f` has a finite-valued diff-Lipschitz constant.  It is thus recommended to set it
    too when instantiating via ``from_source``:

    .. code-block:: python3

       N = 5
       f = from_source(
           cls=pyxu.abc.DiffMap,
           shape=(N, N),
           embed=dict(  # special form to set (diff-)Lipschitz attributes via from_source()
               _diff_lipschitz=2,
           ),
           apply=lambda _, arr: arr**2,
       )
       x = np.arange(N)
       y = f(x)  # [0, 1, 4, 9, 16]
       dL = f.diff_lipschitz  # 2  <- instead of inf
    """
    if embed is None:
        embed = dict()

    if isinstance(vectorize, str):
        vectorize = (vectorize,)
    vectorize = frozenset(vectorize)

    if isinstance(enforce_precision, str):
        enforce_precision = (enforce_precision,)
    enforce_precision = frozenset(enforce_precision)

    src = _FromSource(
        cls=cls,
        shape=shape,
        embed=embed,
        vectorize=vectorize,
        vmethod=vmethod,
        enforce_precision=enforce_precision,
        **kwargs,
    )
    op = src.op()
    return op


class _FromSource:  # See from_source() for a detailed description.
    def __init__(
        self,
        cls: pxt.OpC,
        shape: pxt.OpShape,
        embed: dict,
        vectorize: frozenset[str],
        vmethod: typ.Union[str, dict],
        enforce_precision: frozenset[str],
        **kwargs,
    ):
        from pyxu.abc.operator import _core_operators

        assert cls in _core_operators(), f"Unknown Operator type: {cls}."
        self._op = cls(shape)  # ensure shape well-formed

        # Arithmetic methods to attach to `_op`.
        meth = frozenset.union(*[p.arithmetic_methods() for p in pxa.Property])
        if not (set(kwargs) <= meth):
            msg_head = "Unknown arithmetic methods:"
            unknown = set(kwargs) - meth
            msg_tail = ", ".join([f"{name}()" for name in unknown])
            raise ValueError(f"{msg_head} {msg_tail}")
        self._kwargs = kwargs

        # Extra attributes to attach to `_op`.
        assert isinstance(embed, cabc.Mapping)
        self._embed = embed

        # Add-on functionality to enable.
        self._vkwargs = self._parse_vectorize(vectorize, vmethod)
        self._vectorize = vectorize

        self._ekwargs = self._parse_precision(enforce_precision)
        self._enforce_fp = enforce_precision

    def op(self) -> pxt.OpT:
        _op = self._op  # shorthand
        for p in _op.properties():
            for name in p.arithmetic_methods():
                if func := self._kwargs.get(name, False):
                    # vectorize() & enforce_precision() do NOT kick in for default-provided methods.
                    # (We assume they are Pyxu-compliant from the start.)

                    if name in self._vectorize:
                        decorate = pxu.vectorize(**self._vkwargs[name])
                        func = decorate(func)

                    if name in self._enforce_fp:
                        decorate = pxrt.enforce_precision(**self._ekwargs[name])
                        func = decorate(func)

                    setattr(_op, name, types.MethodType(func, _op))

        # Embed extra attributes
        for name, attr in self._embed.items():
            setattr(_op, name, attr)

        return _op

    def _parse_vectorize(
        self,
        vectorize: frozenset[str],
        vmethod: typ.Union[str, dict],
    ):
        vsize = dict(  # `codim` hint for vectorize()
            apply=self._op.codim,
            prox=self._op.dim,
            grad=self._op.dim,
            adjoint=self._op.dim,
            pinv=self._op.dim,
        )

        if not (vectorize <= set(vsize)):  # un-recognized arithmetic method
            msg_head = "Can only vectorize arithmetic methods"
            msg_tail = ", ".join([f"{name}()" for name in vsize])
            raise ValueError(f"{msg_head} {msg_tail}")

        vmethod_default = pxu.parse_params(
            pxu.vectorize,
            i="bogus",  # doesn't matter
        )["method"]
        if vmethod is None:
            vmethod = vmethod_default
        if isinstance(vmethod, str):
            vmethod = {name: vmethod for name in vsize}

        vkwargs = {
            name: dict(
                i="arr",  # Pyxu arithmetic methods broadcast along parameter `arr`.
                method=vmethod.get(name, vmethod_default),
                codim=vsize[name],
            )
            for name in vsize
        }
        return vkwargs

    def _parse_precision(self, enforce_precision: frozenset[str]):
        ekwargs = dict(
            # Pyxu arithmetic methods enforce FP-precision along these parameters.
            apply=dict(i="arr"),
            prox=dict(i=("arr", "tau")),
            grad=dict(i="arr"),
            adjoint=dict(i="arr"),
            pinv=dict(i=("arr", "damp")),
            svdvals=dict(),
            trace=dict(),
        )

        if not (enforce_precision <= set(ekwargs)):
            msg_head = "Can only enforce precision on arithmetic methods"
            msg_tail = ", ".join([f"{name}()" for name in ekwargs])
            raise ValueError(f"{msg_head} {msg_tail}")
        return ekwargs



import collections.abc as cabc
import warnings

import numpy as np
import pytest

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu


class ViewAs:
    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def func(self) -> cabc.Callable:
        raise NotImplementedError

    @pytest.fixture
    def width_in_out(self) -> tuple[pxrt.Width, pxrt.Width]:
        raise NotImplementedError

    @pytest.fixture(
        params=[
            1,
            (),
            None,
            np.array([1])[0],
        ]
    )
    def non_array_input(self, request):
        return request.param

    @pytest.fixture(
        params=[
            np.bool_,
            np.byte,
            np.ubyte,
            np.short,
            np.ushort,
            np.intc,
            np.uintc,
            np.int_,
            np.uint,
            np.longlong,
            np.ulonglong,
        ]
    )
    def unrecognized_dtype(self, request) -> np.dtype:
        return request.param

    @pytest.fixture
    def no_op_dtype(self) -> np.dtype:
        raise NotImplementedError

    @pytest.fixture
    def valid_data(self) -> tuple[np.ndarray, np.ndarray]:  # input -> output
        # 1D inputs only. Tests needing ND inputs should augment accordingly.
        raise NotImplementedError

    @pytest.fixture
    def _valid_data(self, valid_data, xp, width_in_out):
        return (
            xp.array(valid_data[0], dtype=width_in_out[0].value),
            xp.array(valid_data[1], dtype=width_in_out[1].value),
        )

    # Tests -------------------------------------------------------------------
    def test_fail_non_array_input(self, func, non_array_input):
        with pytest.raises(Exception):
            func(non_array_input)

    def test_fail_unrecognized_dtype(self, func, unrecognized_dtype):
        array = np.arange(50).astype(unrecognized_dtype)
        with pytest.raises(Exception):
            func(array)

    def test_no_op_dtype(self, func, _valid_data, no_op_dtype):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", np.ComplexWarning)
            x = _valid_data[0].astype(no_op_dtype)
        assert x is func(x)

    def test_value1D(self, func, _valid_data):
        in_ = _valid_data[0]
        out_gt = _valid_data[1]
        out = pxu.compute(func(in_))

        assert out.ndim == out_gt.ndim
        assert np.allclose(out, out_gt)

    def test_valueND(self, func, _valid_data):
        sh_extra = (2, 1)  # prepend input/output shape by this amount.

        in_ = _valid_data[0]
        in_ = np.broadcast_to(in_, (*sh_extra, *in_.shape))
        out_gt = _valid_data[1]
        out_gt = np.broadcast_to(out_gt, (*sh_extra, *out_gt.shape))
        out = pxu.compute(func(in_))

        assert out.ndim == out_gt.ndim
        assert np.allclose(out, out_gt)

    def test_backend(self, func, _valid_data):
        out = func(_valid_data[0])
        assert type(out) == type(_valid_data[0])  # noqa: E721

    def test_prec(self, func, _valid_data, width_in_out):
        in_ = _valid_data[0]
        out = func(in_)

        w_in = type(width_in_out[0])
        w_out = type(width_in_out[1])

        assert w_in(in_.dtype).name == w_out(out.dtype).name


class TestViewAsComplex(ViewAs):
    @pytest.fixture
    def func(self) -> cabc.Callable:
        return pxu.view_as_complex

    @pytest.fixture(params=[_.name for _ in pxrt.CWidth])
    def width_in_out(self, request):
        w_in = pxrt.Width[request.param]
        w_out = pxrt.CWidth[request.param]
        return w_in, w_out

    @pytest.fixture
    def valid_data(self):
        in_ = np.arange(6)
        out = np.r_[0, 2, 4] + 1j * np.r_[1, 3, 5]
        return in_, out

    @pytest.fixture(params=[_.value for _ in list(pxrt.CWidth)])
    def no_op_dtype(self, request):
        return request.param


class TestViewAsReal(ViewAs):
    @pytest.fixture
    def func(self) -> cabc.Callable:
        return pxu.view_as_real

    @pytest.fixture(params=[_.name for _ in pxrt.CWidth])
    def width_in_out(self, request):
        w_in = pxrt.CWidth[request.param]
        w_out = pxrt.Width[request.param]
        return w_in, w_out

    @pytest.fixture
    def valid_data(self):
        in_ = np.r_[0, 2, 4] + 1j * np.r_[1, 3, 5]
        out = np.arange(6)
        return in_, out

    @pytest.fixture(params=[_.value for _ in list(pxrt.Width)])
    def no_op_dtype(self, request):
        return request.param


class TestViewAsMat:
    @pytest.fixture(
        params=[
            np.reshape(np.r_[:6] + 1j * np.r_[2:8], (2, 3)),
        ]
    )
    def input(self, request) -> np.ndarray:
        # (M,N) NumPy input to test. Will be transformed to different backend/precisions via _input().
        return request.param

    @pytest.fixture(params=pxd.supported_array_modules())
    def xp(self, request) -> pxt.ArrayModule:
        return request.param

    @pytest.fixture(params=pxrt.CWidth)
    def cwidth(self, request) -> pxrt.CWidth:
        return request.param

    @pytest.fixture
    def width(self, cwidth) -> pxrt.Width:
        return cwidth.real

    @pytest.fixture
    def _input(self, input, xp, cwidth) -> pxt.NDArray:
        in_ = xp.array(input, dtype=cwidth.value)
        return in_

    @pytest.fixture(params=[True, False])
    def real_input(self, request) -> bool:
        return request.param

    @pytest.fixture(params=[True, False])
    def real_output(self, request) -> bool:
        return request.param

    def test_backend_asmat(self, _input, real_input, real_output):
        out = pxu.view_as_real_mat(
            _input,
            real_input=real_input,
            real_output=real_output,
        )
        assert type(out) == type(_input)  # noqa: E721

    def test_prec_asmat(self, _input, real_input, real_output):
        out = pxu.view_as_real_mat(
            _input,
            real_input=real_input,
            real_output=real_output,
        )
        in_prec = pxrt.CWidth(_input.dtype)
        out_prec = pxrt.Width(out.dtype).complex
        assert in_prec == out_prec

    def test_shape_asmat(self, _input, real_input, real_output):
        if real_input and real_output:
            sh_gt = _input.shape
        elif (not real_input) and real_output:
            sh_gt = (_input.shape[0], 2 * _input.shape[1])
        elif real_input and (not real_output):
            sh_gt = (2 * _input.shape[0], _input.shape[1])
        else:  # (not real_input) and (not real_output)
            sh_gt = (2 * _input.shape[0], 2 * _input.shape[1])

        out = pxu.view_as_real_mat(
            _input,
            real_input=real_input,
            real_output=real_output,
        )
        assert out.shape == sh_gt

    def test_math_asmat(self, xp, cwidth, real_input, real_output):
        # view_as_real(A @ x) = view_as_real_mat(A) @ view_as_real(x)
        rng = np.random.default_rng(seed=3)  # random seed for reproducibility

        M, N = 5, 3
        A = xp.array(rng.normal(size=(M, N)) + 1j * rng.normal(size=(M, N)), dtype=cwidth.value)
        x = xp.array(rng.normal(size=(N,)) + 1j * rng.normal(size=(N,)), dtype=cwidth.value)
        if real_input:
            x = x.real

        lhs = pxu.view_as_real(A @ x)
        if real_output:
            lhs = lhs[::2]

        Ar = pxu.view_as_real_mat(A, real_input=real_input, real_output=real_output)
        rhs = Ar @ pxu.view_as_real(x)
        assert np.allclose(lhs, rhs)

    def test_round_trip(self, _input, real_input, real_output):
        # view_as_complex \compose view_as_real = Id
        x = _input
        y = pxu.view_as_real_mat(
            x,
            real_input=real_input,
            real_output=real_output,
        )
        z = pxu.view_as_complex_mat(
            y,
            real_input=real_input,
            real_output=real_output,
        )

        assert np.allclose(x.real, z.real)
        if real_input and real_output:
            assert np.allclose(z.imag, 0)
        else:
            assert np.allclose(x.imag, z.imag)



import collections.abc as cabc
import enum
import importlib.util
import types

import dask.array
import numpy
import packaging.version as pkgv
import scipy.sparse
import sparse

#: Show if CuPy-based backends are available.
CUPY_ENABLED: bool = importlib.util.find_spec("cupy") is not None
if CUPY_ENABLED:
    try:
        import cupy
        import cupyx.scipy.sparse
        import cupyx.scipy.sparse.linalg
    except ImportError:
        # CuPy is installed, but GPU drivers probably missing.
        CUPY_ENABLED = False


@enum.unique
class NDArrayInfo(enum.Enum):
    """
    Supported dense array backends.
    """

    NUMPY = enum.auto()
    DASK = enum.auto()
    CUPY = enum.auto()

    @classmethod
    def default(cls) -> "NDArrayInfo":
        """Default array backend to use."""
        return cls.NUMPY

    def type(self) -> type:
        """Array type associated to a backend."""
        if self.name == "NUMPY":
            return numpy.ndarray
        elif self.name == "DASK":
            return dask.array.core.Array
        elif self.name == "CUPY":
            return cupy.ndarray if CUPY_ENABLED else type(None)
        else:
            raise ValueError(f"No known array type for {self.name}.")

    @classmethod
    def from_obj(cls, obj) -> "NDArrayInfo":
        """Find array backend associated to `obj`."""
        if obj is not None:
            for ndi in cls:
                if isinstance(obj, ndi.type()):
                    return ndi
        raise ValueError(f"No known array type to match {obj}.")

    @classmethod
    def from_flag(cls, gpu: bool) -> "NDArrayInfo":
        """Find array backend suitable for in-memory CPU/GPU computing."""
        if gpu:
            return cls.CUPY
        else:
            return cls.NUMPY

    def module(self, linalg: bool = False) -> types.ModuleType:
        """
        Python module associated to an array backend.

        Parameters
        ----------
        linalg: bool
            Return the linear-algebra submodule with identical API to :py:mod:`numpy.linalg`.
        """
        if self.name == "NUMPY":
            xp = numpy
            xpl = xp.linalg
        elif self.name == "DASK":
            xp = dask.array
            xpl = xp.linalg
        elif self.name == "CUPY":
            xp = cupy if CUPY_ENABLED else None
            xpl = xp if (xp is None) else xp.linalg
        else:
            raise ValueError(f"No known module(s) for {self.name}.")
        return xpl if linalg else xp


@enum.unique
class SparseArrayInfo(enum.Enum):
    """
    Supported sparse array backends.
    """

    SCIPY_SPARSE = enum.auto()
    PYDATA_SPARSE = enum.auto()
    CUPY_SPARSE = enum.auto()

    @classmethod
    def default(cls) -> "SparseArrayInfo":
        """Default array backend to use."""
        return cls.PYDATA_SPARSE

    def type(self) -> type:
        """Array type associated to a backend."""
        if self.name == "SCIPY_SPARSE":
            # All `*matrix` classes descend from `spmatrix`.
            return scipy.sparse.spmatrix
        elif self.name == "PYDATA_SPARSE":
            return sparse.SparseArray
        elif self.name == "CUPY_SPARSE":
            return cupyx.scipy.sparse.spmatrix if CUPY_ENABLED else type(None)
        else:
            raise ValueError(f"No known array type for {self.name}.")

    @classmethod
    def from_obj(cls, obj) -> "SparseArrayInfo":
        """Find array backend associated to `obj`."""
        if obj is not None:
            for sai in cls:
                if isinstance(obj, sai.type()):
                    return sai
        raise ValueError(f"No known array type to match {sai}.")

    def module(self, linalg: bool = False) -> types.ModuleType:
        """
        Python module associated to an array backend.

        Parameters
        ----------
        linalg: bool
            Return the linear-algebra submodule with identical API to :py:mod:`scipy.sparse.linalg`.
        """
        if self.name == "SCIPY_SPARSE":
            xp = scipy.sparse
            xpl = xp.linalg
        elif self.name == "PYDATA_SPARSE":
            xp = sparse
            xpl = scipy.sparse.linalg
        elif self.name == "CUPY_SPARSE":
            xp = cupyx.scipy.sparse if CUPY_ENABLED else None
            xpl = xp if (xp is None) else cupyx.scipy.sparse.linalg
        else:
            raise ValueError(f"No known array module for {self.name}.")
        return xpl if linalg else xp


def supported_array_types() -> cabc.Collection[type]:
    """List of all supported dense array types in current Pyxu install."""
    data = set()
    for ndi in NDArrayInfo:
        if (ndi != NDArrayInfo.CUPY) or CUPY_ENABLED:
            data.add(ndi.type())
    return tuple(data)


def supported_array_modules() -> cabc.Collection[types.ModuleType]:
    """List of all supported dense array modules in current Pyxu install."""
    data = set()
    for ndi in NDArrayInfo:
        if (ndi != NDArrayInfo.CUPY) or CUPY_ENABLED:
            data.add(ndi.module())
    return tuple(data)


def supported_sparse_types() -> cabc.Collection[type]:
    """List of all supported sparse array types in current Pyxu install."""
    data = set()
    for sai in SparseArrayInfo:
        if (sai != SparseArrayInfo.CUPY_SPARSE) or CUPY_ENABLED:
            data.add(sai.type())
    return tuple(data)


def supported_sparse_modules() -> cabc.Collection[types.ModuleType]:
    """List of all supported sparse array modules in current Pyxu install."""
    data = set()
    for sai in SparseArrayInfo:
        if (sai != SparseArrayInfo.CUPY_SPARSE) or CUPY_ENABLED:
            data.add(sai.module())
    return tuple(data)


JAX_SUPPORT = dict(
    min=pkgv.Version("0.4.8"),
    max=pkgv.Version("1.0"),
)
PYTORCH_SUPPORT = dict(
    min=pkgv.Version("2.0"),
    max=pkgv.Version("3.0"),
)

__all__ = [
    "CUPY_ENABLED",
    "NDArrayInfo",
    "SparseArrayInfo",
    "supported_array_types",
    "supported_array_modules",
    "supported_sparse_types",
    "supported_sparse_modules",
    "JAX_SUPPORT",
    "PYTORCH_SUPPORT",
]



import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest


class Median(pxa.Func):
    # f: \bR^{M} -> \bR
    #      x     -> median(x)
    def __init__(self, dim: int):
        super().__init__(shape=(1, dim))

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr):
        xp = pxu.get_array_module(arr)
        y = xp.median(arr, axis=-1, keepdims=True)
        return y


class TestMedian(conftest.FuncT):
    disable_test = frozenset(
        conftest.FuncT.disable_test
        | {
            "test_interface_asloss",  # does not make sense for Median().
        }
    )

    @pytest.fixture(
        params=itertools.product(
            ((5, Median(dim=5)),),
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 test points
            dict(
                in_=dict(arr=np.arange(-5, 6)),
                out=np.array([0]),
            ),
            dict(
                in_=dict(arr=np.arange(200, 350)),
                out=np.array([274.5]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self):
        N_test, N_dim = 5, 3
        return self._random_array((N_test, N_dim))



import warnings

import numpy as np
import scipy.optimize as sopt

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.info.warning as pxw
import pyxu.math as pxm
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "L1Norm",
    "L2Norm",
    "SquaredL2Norm",
    "SquaredL1Norm",
    "LInfinityNorm",
    "L21Norm",
    "PositiveL1Norm",
]


class _ShiftLossMixin:
    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        from pyxu.operator.func.loss import shift_loss

        op = shift_loss(op=self, data=data)
        return op


class L1Norm(_ShiftLossMixin, pxa.ProxFunc):
    r"""
    :math:`\ell_{1}`-norm, :math:`\Vert\mathbf{x}\Vert_{1} := \sum_{i=1}^{N} |x_{i}|`.
    """

    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(1, dim))
        self.lipschitz = np.sqrt(dim)

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        y = pxm.norm(arr, ord=1, axis=-1, keepdims=True)
        return y

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        y = xp.fmax(0, xp.fabs(arr) - tau)
        y *= xp.sign(arr)
        return y


class L2Norm(_ShiftLossMixin, pxa.ProxFunc):
    r"""
    :math:`\ell_{2}`-norm, :math:`\Vert\mathbf{x}\Vert_{2} := \sqrt{\sum_{i=1}^{N} |x_{i}|^{2}}`.
    """

    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(1, dim))
        self.lipschitz = 1
        self.diff_lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        y = pxm.norm(arr, ord=2, axis=-1, keepdims=True)
        return y

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        scale = 1 - tau / xp.fmax(self.apply(arr), tau)

        y = arr.copy()
        y *= scale.astype(dtype=arr.dtype)
        return y


class SquaredL2Norm(pxa.QuadraticFunc):
    r"""
    :math:`\ell^{2}_{2}`-norm, :math:`\Vert\mathbf{x}\Vert^{2}_{2} := \sum_{i=1}^{N} |x_{i}|^{2}`.
    """

    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(1, dim))
        self.lipschitz = np.inf
        self.diff_lipschitz = 2

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        y = pxm.norm(arr, axis=-1, keepdims=True)
        y **= 2
        return y

    @pxrt.enforce_precision(i="arr")
    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        return 2 * arr

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        y = arr.copy()
        y /= 2 * tau + 1
        return y

    def _quad_spec(self):
        from pyxu.operator import HomothetyOp, NullFunc

        Q = HomothetyOp(dim=self.dim, cst=2)
        c = NullFunc(dim=self.dim)
        t = 0
        return (Q, c, t)


class SquaredL1Norm(_ShiftLossMixin, pxa.ProxFunc):
    r"""
    :math:`\ell^{2}_{1}`-norm, :math:`\Vert\mathbf{x}\Vert^{2}_{1} := (\sum_{i=1}^{N} |x_{i}|)^{2}`.
    """

    def __init__(self, dim: pxt.Integer, prox_algo: str = "sort"):
        r"""
        Parameters
        ----------
        dim: Integer
        prox_algo: "root", "sort"
            Algorithm used for computing the proximal operator:

            * 'root' uses [FirstOrd]_ Lemma 6.70,
            * 'sort' uses [OnKerLearn]_ Algorithm 2 (faster).

        Notes
        -----
        Calling :py:meth:`~pyxu.operator.SquaredL1Norm.prox` with DASK inputs when `algo="sort"` is inefficient at
        scale.  Prefer `algo="root"` in this case.
        """
        super().__init__(shape=(1, dim))
        self.lipschitz = np.inf

        algo = prox_algo.strip().lower()
        assert algo in ("root", "sort")
        self._algo = algo

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        y = pxm.norm(arr, ord=1, axis=-1, keepdims=True)
        y **= 2
        return y

    def _prox_root(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        norm = xp.linalg.norm(arr, axis=-1)
        if not (norm > 0):
            out = arr
        else:
            # Part 1: Compute \mu_opt -----------------------------------------
            def f(mu: pxt.Real) -> pxt.Real:
                # Proxy function to compute \mu_opt
                #
                # Inplace implementation of
                #     f(mu) = clip(|arr| * sqrt(tau / mu) - 2 * tau, 0, None).sum() - 1
                x = xp.fabs(arr)
                x *= xp.sqrt(tau / mu)
                x -= 2 * tau
                xp.clip(x, 0, None, out=x)
                y = x.sum()
                y -= 1
                return y

            mu_opt, res = sopt.brentq(
                f=f,
                a=1e-12,
                b=xp.fabs(arr).max() ** 2 / (4 * tau),
                full_output=True,
                disp=False,
            )
            if not res.converged:
                msg = "Computing mu_opt did not converge."
                raise ValueError(msg)

            # Part 2: Compute \lambda -----------------------------------------
            # Inplace implementation of
            #     lambda_ = clip(|arr| * sqrt(tau / mu_opt) - 2 * tau, 0, None)
            lambda_ = xp.fabs(arr)
            lambda_ *= xp.sqrt(tau / mu_opt)
            lambda_ -= 2 * tau
            xp.clip(lambda_, 0, None, out=lambda_)

            # Part 3: Compute \prox -------------------------------------------
            # Inplace implementation of
            #     out = arr * lambda_ / (lambda_ + 2 * tau)
            out = arr.copy()
            out *= lambda_
            lambda_ += 2 * tau
            out /= lambda_
        return out

    def _prox_sort(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)

        # Inplace implementation of
        #     y = sort(|arr|, axis=-1)[::-1]
        y = xp.fabs(arr)
        y.sort(axis=-1)
        y = y[::-1]

        z = y.cumsum(axis=-1)
        z *= tau / (0.5 + tau * xp.arange(1, arr.size + 1, dtype=z.dtype))
        tau2 = z[max(xp.flatnonzero(y > z), default=0)]

        out = L1Norm(dim=self.dim).prox(arr, tau2)
        return out

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        N = pxd.NDArrayInfo
        if (N.from_obj(arr) == N.DASK) and (self._algo == "sort"):
            msg = "\n".join(
                [
                    "Using prox_algo='sort' on DASK inputs is inefficient.",
                    "Consider using prox_algo='root' instead.",
                ]
            )
            warnings.warn(msg, pxw.PerformanceWarning)

        vectorize = pxu.vectorize(
            # DASK backend required since prox_[root|sort]() don't accept DASK inputs.
            i="arr",
            method="parallel",
            codim=arr.shape[-1],
        )
        f = dict(
            root=vectorize(self._prox_root),
            sort=vectorize(self._prox_sort),
        )[self._algo]

        out = f(arr, tau)
        return out


class LInfinityNorm(_ShiftLossMixin, pxa.ProxFunc):
    r"""
    :math:`\ell_{\infty}`-norm, :math:`\Vert\mathbf{x}\Vert_{\infty} := \max_{i=1,\ldots,N} |x_{i}|`.
    """

    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(1, dim))
        self.lipschitz = 1

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        y = pxm.norm(arr, ord=np.inf, axis=-1, keepdims=True)
        return y

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        vectorize = pxu.vectorize(
            i="arr",
            method="parallel",
            codim=arr.shape[-1],
        )
        f = vectorize(self._prox)

        out = f(arr, tau)
        return out

    def _prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        y = xp.zeros(arr.shape, dtype=arr.dtype)

        def f(mu: pxt.Real) -> pxt.Real:
            # Proxy function to compute \mu_opt
            #
            # Inplace implementation of
            #     f(mu) = clip(xp.fabs(arr) - mu, 0, None).sum() - tau
            x = xp.fabs(arr)
            x -= mu
            xp.clip(x, 0, None, out=x)
            y = x.sum()
            y -= tau
            return y

        mu_max = self.apply(arr)
        if (mu_max > 0).all():
            mu_opt = sopt.brentq(f, a=0, b=mu_max)

            # Inplace implementation of
            #     y = sgn(arr) * fmin(|arr|, mu_opt)
            xp.fabs(arr, out=y)
            xp.fmin(y, mu_opt, out=y)
            y *= xp.sign(arr)

        return y


class L21Norm(_ShiftLossMixin, pxa.ProxFunc):
    r"""
    Mixed :math:`\ell_{2}-\ell_{1}` norm, :math:`\Vert\mathbf{x}\Vert_{2, 1} := \sum_{i=1}^{N} \sqrt{\sum_{j=1}^{M}
    x_{i, j}^{2}}`.

    Note
    ----
    The input array need not be 2-dimensional: the :math:`\ell_{2}` norm is applied along a predefined subset of
    dimensions, and the :math:`\ell_{1}` norm on the remaining ones.
    """

    def __init__(
        self,
        arg_shape: pxt.NDArrayShape,
        l2_axis: pxt.NDArrayAxis = (0,),
    ):
        r"""
        Parameters
        ----------
        arg_shape: NDArrayShape
            Shape of the input array.
        l2_axis: NDArrayAxis
            Axis (or axes) along which the :math:`\ell_{2}` norm is applied.
        """
        arg_shape = pxu.as_canonical_shape(arg_shape)
        assert all(a > 0 for a in arg_shape)
        N_dim = len(arg_shape)
        assert N_dim >= 2

        l2_axis = np.unique(pxu.as_canonical_shape(l2_axis))  # drop potential duplicates
        assert np.all((-N_dim <= l2_axis) & (l2_axis < N_dim))  # all axes in valid range
        l2_axis = (l2_axis + N_dim) % N_dim  # get rid of negative axes
        l1_axis = np.setdiff1d(np.arange(N_dim), l2_axis)

        super().__init__(shape=(1, np.prod(arg_shape)))
        self.lipschitz = np.inf
        self._arg_shape = arg_shape
        self._l1_axis = l1_axis
        self._l2_axis = l2_axis

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray):
        sh = arr.shape[:-1]
        arr = arr.reshape(sh + self._arg_shape)

        l2_axis = tuple(len(sh) + self._l2_axis)
        x = (arr**2).sum(axis=l2_axis, keepdims=True)
        xp = pxu.get_array_module(arr)
        xp.sqrt(x, out=x)

        l1_axis = tuple(len(sh) + self._l1_axis)
        out = x.sum(axis=l1_axis, keepdims=True)
        return out.reshape(*sh, -1)

    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real):
        sh = arr.shape[:-1]
        arr = arr.reshape(sh + self._arg_shape)

        l2_axis = tuple(len(sh) + self._l2_axis)
        n = (arr**2).sum(axis=l2_axis, keepdims=True)
        xp = pxu.get_array_module(arr)
        xp.sqrt(n, out=n)

        out = arr.copy()
        out *= 1 - tau / xp.fmax(n, tau)
        return out.reshape(*sh, -1)


class PositiveL1Norm(_ShiftLossMixin, pxa.ProxFunc):
    r"""
    :math:`\ell_{1}`-norm, with a positivity constraint.

    .. math::

       f(\mathbf{x})
       :=
       \lVert\mathbf{x}\rVert_{1} + \iota_{+}(\mathbf{x}),

    .. math::

       \textbf{prox}_{\tau f}(\mathbf{z})
       :=
       \max(\mathrm{soft}_\tau(\mathbf{z}), \mathbf{0})

    See Also
    --------
    :py:class:`~pyxu.operator.PositiveOrthant`
    """

    def __init__(self, dim: pxt.Integer):
        super().__init__(shape=(1, dim))
        from pyxu.operator.func.indicator import PositiveOrthant

        self._indicator = PositiveOrthant(dim=dim)
        self._l1norm = L1Norm(dim=dim)
        self.lipschitz = np.inf

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        return self._indicator(arr) + self._l1norm(arr)

    @pxrt.enforce_precision(i=["arr", "tau"])
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        xp = pxu.get_array_module(arr)
        return xp.fmax(0, arr - tau)



import collections.abc as cabc

import numpy as np
import pytest
import scipy.ndimage as scimage

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest

try:
    import scipy.ndimage._filters as scif
except ImportError:
    import scipy.ndimage.filters as scif


def diff_params_fd(scheme, accuracy, sampling):  # Finite Difference
    diff_params = {
        "scheme": scheme,
        "accuracy": accuracy,
    }
    gt_diff = {
        # diff type
        "central": {
            # accuracy
            2: {
                # order
                0: {
                    "coefs": np.array([1]),
                    "origin": 0,
                },
                1: {
                    "coefs": np.array([-1 / 2, 0, 1 / 2]) / sampling,
                    "origin": 1,
                },
                2: {
                    "coefs": np.array([1, -2, 1]) / (sampling**2),
                    "origin": 1,
                },
            },
            4: {
                0: {
                    "coefs": np.array([1]),
                    "origin": 0,
                },
                1: {
                    "coefs": np.array([1 / 12, -2 / 3, 0, 2 / 3, -1 / 12]) / sampling,
                    "origin": 2,
                },
                2: {
                    "coefs": np.array([-1 / 12, 4 / 3, -5 / 2, 4 / 3, -1 / 12]) / (sampling**2),
                    "origin": 2,
                },
            },
        },
        "forward": {
            2: {
                0: {
                    "coefs": np.array([1]),
                    "origin": 0,
                },
                1: {
                    "coefs": np.array([-3 / 2, 2, -1 / 2]) / sampling,
                    "origin": 0,
                },
                2: {
                    "coefs": np.array([2, -5, 4, -1]) / (sampling**2),
                    "origin": 0,
                },
            },
            4: {
                0: {
                    "coefs": np.array([1]),
                    "origin": 0,
                },
                1: {
                    "coefs": np.array([-25 / 12, 4, -3, 4 / 3, -1 / 4]) / sampling,
                    "origin": 0,
                },
                2: {
                    "coefs": np.array([15 / 4, -77 / 6, 107 / 6, -13, 61 / 12, -5 / 6]) / (sampling**2),
                    "origin": 0,
                },
            },
        },
        "backward": {
            2: {
                0: {
                    "coefs": np.array([1]),
                    "origin": 0,
                },
                1: {
                    "coefs": np.array([1 / 2, -2, 3 / 2]) / sampling,
                    "origin": 2,
                },
                2: {
                    "coefs": np.array([-1, 4, -5, 2]) / (sampling**2),
                    "origin": 3,
                },
            }
        },
    }
    gt_diffs = gt_diff[scheme][accuracy]

    return diff_params, gt_diffs


def diff_params_gd(sigma, truncate, sampling):  # Gaussian Derivative
    diff_params = {"sigma": sigma, "truncate": truncate}

    sigma_pix = sigma / sampling  # Sigma rescaled to pixel units
    radius = int(truncate * float(sigma_pix) + 0.5)
    gt_diffs = {
        0: {
            "coefs": np.flip(scif._gaussian_kernel1d(sigma_pix, 0, radius)),
            "origin": radius,
        },
        1: {
            "coefs": np.flip(scif._gaussian_kernel1d(sigma_pix, 1, radius)) / sampling,
            "origin": radius,
        },
        2: {
            "coefs": np.flip(scif._gaussian_kernel1d(sigma_pix, 2, radius)) / (sampling**2),
            "origin": radius,
        },
    }
    return diff_params, gt_diffs


def apply_derivative(arr, arg_shape, axis, gt_diffs, order, mode="constant"):
    # Apply derivative based on ground truth kernel

    axis = len(arg_shape) if axis == -1 else axis
    coefs = gt_diffs[order]["coefs"]
    origin = np.zeros(len(arg_shape) + 1, dtype="int8")
    origin[axis + 1] = gt_diffs[order]["origin"] - (len(coefs) // 2)
    kernel = np.array(coefs).reshape(*((1,) * len(arg_shape)), -1).swapaxes(axis + 1, -1)

    # Scipy and numpy padding modes have different names.
    mode = mode if mode != "reflect" else "mirror"
    mode = mode if mode != "symmetric" else "reflect"
    mode = mode if mode != "edge" else "nearest"

    return scimage.correlate(
        arr.reshape(-1, *arg_shape),
        kernel,
        mode=mode,
        origin=origin,
        cval=0.0,
    )


def apply_gradient(arr, arg_shape, gt_diffs, directions, diff_method, mode="constant"):
    if diff_method == "fd":
        pd = [apply_derivative(arr, arg_shape, axis, gt_diffs, 1, mode) for axis in directions]
    else:
        # diff_method == "gd"
        pd = []
        for axis in directions:
            out = apply_derivative(arr, arg_shape, axis, gt_diffs, 1, mode)
            # for smooth_axis in directions:
            for smooth_axis in range(len(arg_shape)):
                if smooth_axis != axis:
                    out = apply_derivative(out, arg_shape, smooth_axis, gt_diffs, 0, mode)
            pd.append(out)
    return np.stack(pd)


def apply_hessian(arr, arg_shape, gt_diffs, directions, diff_method, mode="constant"):
    pd = []
    for ax1, ax2 in directions:
        if ax1 == ax2:
            out = apply_derivative(arr, arg_shape, ax1, gt_diffs, 2, mode)
        else:
            out = apply_derivative(arr, arg_shape, ax1, gt_diffs, 1, mode)
            out = apply_derivative(out, arg_shape, ax2, gt_diffs, 1, mode)

        if diff_method == "gd":
            # for smooth_axis in directions:
            for smooth_axis in range(len(arg_shape)):
                if smooth_axis not in [ax1, ax2]:
                    out = apply_derivative(out, arg_shape, smooth_axis, gt_diffs, 0, mode)
        pd.append(out)
    return np.stack(pd)


@pytest.mark.filterwarnings("ignore::pyxu.info.warning.PrecisionWarning")
@pytest.mark.filterwarnings("ignore::numba.core.errors.NumbaPerformanceWarning")
class DiffOpMixin(conftest.LinOpT):
    @pytest.fixture(
        params=[
            # 1,
            1.5,
            # 2
        ]
    )
    def sampling(self, request):
        return request.param

    @pytest.fixture(params=pxd.NDArrayInfo)
    def ndi(self, request):
        # [Sepand] Not inferred from spec(diff_kwargs) [unlike most Pyxu operators] since diff_kwargs() needs this information beforehand.
        ndi_ = request.param
        if ndi_.module() is None:
            pytest.skip(f"{ndi_} unsupported on this machine.")
        return ndi_

    @pytest.fixture(params=pxrt.Width)
    def width(self, request):
        # [Sepand] Not inferred from spec(diff_kwargs) [unlike most Pyxu operators] since diff_kwargs() needs this information beforehand.
        return request.param

    @pytest.fixture(params=["fd", "gd"])
    def diff_method(self, request):
        return request.param

    @pytest.fixture(
        params=[
            #  Finite Diff. ,   Gaussian Der.
            # (diff_typ, acc), (sigma, truncate)
            (("central", 2), (1.0, 1.0)),
        ]
    )
    def init_params(self, diff_method, sampling, request):
        params_fd, params_gd = request.param
        if diff_method == "fd":
            return diff_params_fd(params_fd[0], params_fd[1], sampling)
        elif diff_method == "gd":
            return diff_params_gd(params_gd[0], params_gd[1], sampling)

    @pytest.fixture
    def diff_params(self, init_params):
        return init_params[0]

    @pytest.fixture
    def gt_diffs(self, init_params):
        return init_params[1]

    @pytest.fixture
    def diff_op(self) -> pxt.OpC:
        # To override in subclasses.
        raise NotImplementedError

    @pytest.fixture
    def diff_kwargs(self) -> dict:
        # To override in subclasses.
        raise NotImplementedError

    @pytest.fixture
    def spec(
        self,
        diff_op,
        diff_params,
        diff_kwargs,
        ndi,
        width,
    ) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        kwargs = diff_params.copy()
        kwargs.update(diff_kwargs)
        with pxrt.Precision(width):
            op = diff_op(**kwargs)
        return op, ndi, width


class TestPartialDerivative(DiffOpMixin):
    @pytest.fixture(
        params=[
            # (arg_shape, order, mode)
            ((10,), (1,), "constant"),
            ((10, 10), (2, 1), ("edge", "constant")),
            (
                (10, 10),
                (
                    0,
                    1,
                ),
                "edge",
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, order, mode) configs to test
        # * `request.param[0]` corresponds to raw inputs users provide to DiffOp().
        # * `request.param[1]` corresponds to their ground-truth canonical parameterization.
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):  # canonical representation
        arg_shape, _, _ = _spec
        return arg_shape

    @pytest.fixture
    def order(self, _spec):  # canonical representation (NumPy)
        _, order, _ = _spec
        return order

    @pytest.fixture
    def mode(self, _spec):  # canonical representation
        _, _, mode = _spec
        return mode

    @pytest.fixture
    def data_shape(self, arg_shape) -> pxt.NDArrayShape:
        size = np.prod(arg_shape).item()
        sh = (size, size)
        return sh

    @pytest.fixture(
        params=[
            #  Finite Diff. ,   Gaussian Der.
            # (diff_typ, acc), (sigma, truncate)
            (("forward", 2), (2.0, 1.0)),
            (("forward", 4), (1.0, 1.0)),
            (("backward", 2), (1.0, 1.0)),
            (("central", 2), (1.0, 1.0)),
            (("central", 4), (1.0, 1.0)),
        ]
    )
    def init_params(self, diff_method, sampling, request):
        params_fd, params_gd = request.param
        if diff_method == "fd":
            return diff_params_fd(params_fd[0], params_fd[1], sampling)
        elif diff_method == "gd":
            return diff_params_gd(params_gd[0], params_gd[1], sampling)

    @pytest.fixture
    def diff_kwargs(self, order, arg_shape, mode, ndi, width, sampling):
        return {
            "order": order,
            "arg_shape": arg_shape,
            "mode": mode,
            "gpu": ndi == pxd.NDArrayInfo.CUPY,
            "dtype": width.value,
            "sampling": sampling,
        }

    @pytest.fixture
    def data_apply(self, op, gt_diffs, order, arg_shape, mode) -> conftest.DataLike:
        arr = self._random_array((op.dim,), seed=20)

        order = (order,) if not isinstance(order, tuple) else order
        mode = (mode,) * len(arg_shape) if not isinstance(mode, tuple) else mode

        out = arr.copy()
        for ax, ord_ in enumerate(order):
            out = apply_derivative(out, arg_shape, ax, gt_diffs, ord_, mode[ax])

        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )

    @pytest.fixture
    def diff_op(self, diff_method):
        if diff_method == "fd":
            return pxo.PartialDerivative.finite_difference
        elif diff_method == "gd":
            return pxo.PartialDerivative.gaussian_derivative


class TestGradient(DiffOpMixin):
    @pytest.fixture(
        params=[
            #          (arg_shape, directions)
            (
                (5,),
                (0,),
            ),
            (
                (5, 5),
                (0, 1),
            ),
            (
                (5, 5),
                None,
            ),
            (
                (5, 5, 5),
                (0, 2),
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, directions) configs to test
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):
        return _spec[0]

    @pytest.fixture
    def directions(self, _spec):
        return _spec[1]

    @pytest.fixture
    def data_shape(self, arg_shape, directions) -> pxt.NDArrayShape:
        size = np.prod(arg_shape).item()
        n_derivatives = len(directions) if directions is not None else len(arg_shape)
        sh = (size * n_derivatives, size)
        return sh

    @pytest.fixture
    def diff_op(self):
        return pxo.Gradient

    @pytest.fixture()
    def diff_kwargs(self, arg_shape, directions, ndi, width, sampling, diff_method):
        return {
            "arg_shape": arg_shape,
            "directions": directions,
            "mode": "constant",
            "gpu": ndi == pxd.NDArrayInfo.CUPY,
            "dtype": width.value,
            "sampling": sampling,
            "diff_method": diff_method,
        }

    @pytest.fixture
    def data_apply(self, op, arg_shape, diff_method, gt_diffs, directions) -> conftest.DataLike:
        arr = self._random_array(arg_shape, seed=20)
        directions = np.arange(len(arg_shape)) if directions is None else directions
        out = apply_gradient(
            arr,
            arg_shape=arg_shape,
            gt_diffs=gt_diffs,
            directions=directions,
            diff_method=diff_method,
        )
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestHessian(DiffOpMixin):
    @pytest.fixture(
        params=[
            #       (arg_shape, directions)
            #       (canonical directions)
            # 1D
            (((5,), 0), ((0, 0),)),  # Mode 0 (See Hessian Notes)
            (((5,), (0, 0)), ((0, 0),)),  # Mode 1
            (((5,), ((0, 0),)), ((0, 0),)),  # Mode 2
            (((5,), "all"), ((0, 0),)),  # Mode 3
            # ND
            (((5, 5, 5), 1), ((1, 1),)),  # Mode 0
            (((5, 5, 5), (0, 2)), ((0, 2),)),  # Mode 1
            (((5, 5, 5), ((1, 1), (1, 2))), ((1, 1), (1, 2))),  # Mode 2
            (((5, 5, 5), "all"), ((0, 0), (0, 1), (0, 2), (1, 1), (1, 2), (2, 2))),  # Mode 3
        ]
    )
    def _spec(self, request):
        # (arg_shape, directions) configs to test
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def directions(self, _spec):
        return _spec[0][1]

    @pytest.fixture
    def canonical_directions(self, _spec):
        return _spec[1]

    @pytest.fixture
    def data_shape(self, arg_shape, directions) -> pxt.NDArrayShape:
        size = np.prod(arg_shape).item()
        if isinstance(directions, int):  # Case 0
            sh = (size, size)
        elif isinstance(directions, cabc.Sequence):
            if isinstance(directions, str):  # case 3
                n_derivatives = len(arg_shape) * (len(arg_shape) + 1) // 2
                sh = (size * n_derivatives, size)
            elif isinstance(directions[0], int):  # Case 1
                sh = (size, size)
            elif isinstance(directions[0], cabc.Sequence):  # Case 2
                sh = (size * len(directions), size)
        return sh

    @pytest.fixture
    def diff_op(self):
        return pxo.Hessian

    @pytest.fixture
    def diff_kwargs(self, arg_shape, directions, ndi, width, sampling, diff_method):
        return {
            "arg_shape": arg_shape,
            "directions": directions,
            "mode": "constant",
            "gpu": ndi == pxd.NDArrayInfo.CUPY,
            "dtype": width.value,
            "sampling": sampling,
            "diff_method": diff_method,
        }

    @pytest.fixture
    def data_apply(self, op, arg_shape, diff_method, gt_diffs, canonical_directions) -> conftest.DataLike:
        arr = self._random_array(arg_shape, seed=20)  # random seed for reproducibility

        out = apply_hessian(
            arr,
            arg_shape=arg_shape,
            gt_diffs=gt_diffs,
            directions=canonical_directions,
            diff_method=diff_method,
        )
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestJacobian(DiffOpMixin):
    @pytest.fixture(
        params=[
            #          (arg_shape, directions)
            (
                (5,),
                (0,),
            ),
            (
                (5, 5),
                (0, 1),
            ),
            (
                (5, 5),
                None,
            ),
            (
                (5, 5, 5),
                (0, 2),
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, directions) configs to test
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):
        return _spec[0]

    @pytest.fixture
    def directions(self, _spec):
        return _spec[1]

    @pytest.fixture(params=[1, 3])
    def n_channels(self, request):
        return request.param

    @pytest.fixture
    def data_shape(self, arg_shape, n_channels, directions) -> pxt.NDArrayShape:
        size = np.prod(arg_shape).item()
        n_derivatives = len(directions) if directions is not None else len(arg_shape)
        sh = (size * n_derivatives * n_channels, size * n_channels)
        return sh

    @pytest.fixture
    def diff_op(self):
        return pxo.Jacobian

    @pytest.fixture
    def diff_kwargs(self, arg_shape, n_channels, directions, ndi, width, sampling, diff_method):
        return {
            "arg_shape": arg_shape,
            "n_channels": n_channels,
            "directions": directions,
            "mode": "constant",
            "gpu": ndi == pxd.NDArrayInfo.CUPY,
            "dtype": width.value,
            "sampling": sampling,
            "diff_method": diff_method,
        }

    @pytest.fixture
    def data_apply(self, op, arg_shape, n_channels, diff_method, gt_diffs, directions) -> conftest.DataLike:
        arr = self._random_array((n_channels,) + arg_shape, seed=20)  # random seed for reproducibility
        directions = np.arange(len(arg_shape)) if directions is None else directions
        out = []
        for ch in range(n_channels):
            out.append(
                apply_gradient(
                    arr[ch],
                    arg_shape=arg_shape,
                    gt_diffs=gt_diffs,
                    directions=directions,
                    diff_method=diff_method,
                )
            )
        out = np.concatenate(out)
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestDivergence(DiffOpMixin):
    @pytest.fixture(
        params=[
            #          (arg_shape, directions)
            (
                (5,),
                (0,),
            ),
            (
                (5, 5),
                (0, 1),
            ),
            (
                (5, 5),
                None,
            ),
            (
                (5, 5, 5),
                (0, 2),
            ),
        ]
    )
    def _spec(self, request):
        # (arg_shape, directions) configs to test
        return request.param

    @pytest.fixture
    def arg_shape(self, _spec):
        return _spec[0]

    @pytest.fixture
    def directions(self, _spec):
        return _spec[1]

    @pytest.fixture
    def data_shape(self, arg_shape, directions) -> pxt.NDArrayShape:
        size = np.prod(arg_shape).item()
        n_derivatives = len(directions) if directions is not None else len(arg_shape)
        sh = (size, size * n_derivatives)
        return sh

    @pytest.fixture
    def diff_op(self):
        return pxo.Divergence

    @pytest.fixture
    def diff_kwargs(self, arg_shape, directions, ndi, width, sampling, diff_method):
        return {
            "arg_shape": arg_shape,
            "directions": directions,
            "mode": "constant",
            "gpu": ndi == pxd.NDArrayInfo.CUPY,
            "dtype": width.value,
            "sampling": sampling,
            "diff_method": diff_method,
        }

    @pytest.fixture
    def data_apply(self, op, arg_shape, gt_diffs, directions, diff_method) -> conftest.DataLike:
        directions = np.arange(len(arg_shape)) if directions is None else directions
        arr = self._random_array((len(directions),) + arg_shape, seed=20)
        out = [
            apply_gradient(
                arr[i],
                arg_shape=arg_shape,
                gt_diffs=gt_diffs,
                directions=(ax,),
                diff_method=diff_method,
            )
            for i, ax in enumerate(directions)
        ]
        out = np.stack(out).sum(axis=0)
        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )


class TestLaplacian(DiffOpMixin):
    @pytest.fixture(params=[(5,), (5, 5, 5)])
    def arg_shape(self, request):
        return request.param

    @pytest.fixture
    def data_shape(self, arg_shape) -> pxt.NDArrayShape:
        size = np.prod(arg_shape).item()
        sh = (size, size)
        return sh

    @pytest.fixture
    def diff_op(self):
        return pxo.Laplacian

    @pytest.fixture
    def diff_kwargs(self, arg_shape, ndi, width, sampling, diff_method):
        return {
            "arg_shape": arg_shape,
            "mode": "constant",
            "gpu": ndi == pxd.NDArrayInfo.CUPY,
            "dtype": width.value,
            "sampling": sampling,
            "diff_method": diff_method,
        }

    @pytest.fixture
    def data_apply(self, op, arg_shape, gt_diffs, diff_method) -> conftest.DataLike:
        arr = self._random_array(arg_shape, seed=20)  # random seed for reproducibility
        directions = tuple([(i, i) for i in range(len(arg_shape))])
        out = apply_hessian(
            arr,
            arg_shape=arg_shape,
            gt_diffs=gt_diffs,
            directions=directions,
            diff_method=diff_method,
        ).sum(axis=0)

        return dict(
            in_=dict(arr=arr.reshape(-1)),
            out=out.reshape(-1),
        )
